from .. import get_python_manager, none, Some, \
    get_remote_manager, get_mode, _get_source_table
from .loaders import MetaLoaderFile, Materializations
from .DataStructures.RegField import RegField
from .DataStructures.MetaField import MetaField
from .DataStructures.Aggregates import *
from .DataStructures.GenometricPredicates import GenometricCondition
from .DataStructures import reg_fixed_fileds
from .import GDataframe
from .loaders import MemoryLoader, MetadataProfiler
from ..FileManagment.TempFileManager import get_unique_identifier, get_new_dataset_tmp_folder
from .loaders.Sources import PARSER, LOCAL, REMOTE
from .storers.parserToXML import parserToXML
import os


class GMQLDataset:
    """
    The main abstraction of the library.
    A GMQLDataset represents a genomic dataset in the GMQL standard and it is divided
    in region data and meta data. The function that can be applied to a GMQLDataset
    affect one of these two features or both.
    """

    def __init__(self, parser=None, index=None, location="local", path_or_name=None,
                 local_sources=None, remote_sources=None,
                 meta_profile=None):
        self.parser = parser
        self.__index = index
        self.location = location
        self.meta_profile = None
        if isinstance(meta_profile, MetadataProfiler.MetadataProfile):
            self.meta_profile = meta_profile
        elif meta_profile is not None:
            raise TypeError("meta_profile must be MetadataProfiler")
        self.path_or_name = path_or_name
        self.pmg = get_python_manager()
        self.opmng = self.pmg.getOperatorManager()

        # provenance
        if isinstance(local_sources, list):
            self._local_sources = local_sources
        else:
            self._local_sources = []
        if isinstance(remote_sources, list):
            self._remote_sources = remote_sources
        else:
            self._remote_sources = []

        # get the schema of the dataset
        schemaJava = self.pmg.getVariableSchemaNames(self.__index)
        self.schema = []
        for field in schemaJava:
            self.schema.append(field)
        self.schema.extend(reg_fixed_fileds)

        # setting the schema as properties of the dataset
        for field in self.schema:
            self.__setattr__(field, self.RegField(field))
        # add also left and right
        self.left = self.RegField("left")
        self.right = self.RegField("right")

    def __getitem__(self, arg):
        if isinstance(arg, tuple) and len(arg) == 2:
            item, t = arg
            return self.MetaField(name=item, t=t)
        elif isinstance(arg, str):
            return self.MetaField(name=arg)
        elif isinstance(arg, MetaField):
            return self.meta_select(predicate=arg)
        else:
            raise ValueError("Input must be a string or a MetaField. {} was found".format(type(arg)))

    def __show_info(self):
        print("GMQLDataset")
        print("\tParser:\t{}".format(self.parser.get_parser_name()))
        print("\tIndex:\t{}".format(self.__index))

    def get_metadata(self):
        if self.path_or_name is None:
            raise ValueError("You cannot explore the metadata of an intermediate query."
                             "You can get metadata only after a load_from_local or load_from_remote")
        if self.location == 'local':
            return self.__get_metadata_local()
        elif self.location == 'remote':
            return self.__get_metadata_remote()

    def __get_metadata_local(self):
        meta = MetaLoaderFile.load_meta_from_path(self.path_or_name)
        return meta

    def __get_metadata_remote(self):
        pass

    def get_reg_attributes(self):
        """
        Returns the region fields of the dataset
        
        :return: a list of field names
        """
        return self.schema

    def MetaField(self, name, t=None):
        """
        Creates an instance of a metadata field of the dataset. It can be used in building expressions
        or conditions for projection or selection.
        Notice that this function is equivalent to call::
            
            dataset[name]
        
        :param name: the name of the metadata that is considered
        :param t: the type of the metadata attribute

        :return: a MetaField instance
        """
        return MetaField(name=name, index=self.__index, t=t)

    def RegField(self, name):
        """
        Creates an instance of a region field of the dataset. It can be used in building expressions
        or conditions for projection or selection.
        Notice that this function is equivalent to::
        
            dataset.name
        
        :param name: the name of the region field that is considered
        :return: a RegField instance
        """
        return RegField(name=name, index=self.__index)

    def select(self, meta_predicate=None, region_predicate=None,
               semiJoinDataset=None,  semiJoinMeta=None):

        semiJoinDataset_exists = False
        if isinstance(meta_predicate, MetaField):
            meta_condition = Some(meta_predicate.getMetaCondition())
        elif meta_predicate is None:
            meta_condition = none()
        else:
            raise TypeError("meta_predicate must be a MetaField or None."
                            " {} was provided".format(type(meta_predicate)))

        if isinstance(region_predicate, RegField):
            region_condition = Some(region_predicate.getRegionCondition())
        elif region_predicate is None:
            region_condition = none()
        else:
            raise TypeError("region_predicate must be a RegField or None."
                            " {} was provided".format(type(region_predicate)))

        if isinstance(semiJoinDataset, GMQLDataset):
            other_dataset = Some(semiJoinDataset.__index)
            semiJoinDataset_exists = True
        elif semiJoinDataset is None:
            other_dataset = none()
        else:
            raise TypeError("semiJoinDataset must be a GMQLDataset or None."
                            " {} was provided".format(type(semiJoinDataset)))

        if isinstance(semiJoinMeta, list) and \
                all([isinstance(x, str) for x in semiJoinMeta]):
            if semiJoinDataset_exists:
                semi_join = Some(self.opmng.getMetaJoinCondition(semiJoinMeta))
            else:
                raise ValueError("semiJoinDataset and semiJoinMeta must be present at the "
                                 "same time or totally absent")
        elif semiJoinMeta is None:
            semi_join = none()
        else:
            raise TypeError("semiJoinMeta must be a list of strings or None."
                            " {} was provided".format(type(semiJoinMeta)))

        new_index = self.opmng.select(self.__index, other_dataset,
                                      semi_join, meta_condition, region_condition)
        return GMQLDataset(index=new_index, location=self.location, local_sources=self._local_sources,
                           remote_sources=self._remote_sources, meta_profile=self.meta_profile)

    def meta_select(self, predicate=None, semiJoinDataset=None, semiJoinMeta=None):
        """
        The META_SELECT operation creates a new dataset from an existing one 
        by extracting a subset of samples from the input dataset; each sample 
        in the output dataset has the same region attributes and metadata 
        as in the input dataset.
        The selection can be based on:
        
            * *Metadata predicates*: selection based on the existence and values of certain 
              metadata attributes in each sample. In predicates, attribute-value conditions 
              can be composed using logical predicates & (and), | (or) and ~ (not)
            * *SemiJoin clauses*: selection based on the existence of certain metadata :attr:`~.semiJoinMeta`
              attributes and the matching of their values with those associated with at 
              least one sample in an external dataset :attr:`~.semiJoinDataset`
        
        :param predicate: logical predicate on the values of the rows
        :param semiJoinDataset: an other GMQLDataset 
        :param semiJoinMeta: a list of metadata
        :return: a new GMQLDataset

        
        Example 1::
        
            output_dataset = dataset.meta_select(dataset['patient_age'] < 70)
            # This statement can be written also as
            output_dataset = dataset[ dataset['patient_age'] < 70 ]

        Example 2::

            output_dataset = dataset.meta_select( (dataset['tissue_status'] == 'tumoral') &
                                                (tumor_tag != 'gbm') | (tumor_tag == 'brca'))
            # This statement can be written also as
            output_dataset = dataset[ (dataset['tissue_status'] == 'tumoral') &
                                    (tumor_tag != 'gbm') | (tumor_tag == 'brca') ]

        Example 3::

            JUN_POLR2A_TF = HG19_ENCODE_NARROW.meta_select( JUN_POLR2A_TF['antibody_target'] == 'JUN',
                                                            semiJoinDataset=POLR2A_TF, semiJoinMeta=['cell'])

        The meta selection predicate can use all the classical equalities and disequalities
        {>, <, >=, <=, ==, !=} and predicates can be connected by the classical logical symbols
        {& (AND), | (OR), ~ (NOT)} plus the *isin* function.

        """

        return self.select(meta_predicate=predicate, semiJoinDataset=semiJoinDataset, semiJoinMeta=semiJoinMeta)

    def reg_select(self, predicate=None, semiJoinDataset=None, semiJoinMeta=None):
        """
        Select only the regions in the dataset that satisfy the predicate

        :param predicate: logical predicate on the values of the regions
        :param semiJoinDataset: an other GMQLDataset 
        :param semiJoinMeta: a list of metadata
        :return: a new GMQLDataset

        An example of usage::

            new_dataset = dataset.reg_select((dataset.chr == 'chr1') | (dataset.pValue < 0.9))

        You can also use Metadata attributes in selection::

            new_dataset = dataset.reg_select(dataset.score > dataset['size'])

        This statement selects all the regions whose field score is strictly higher than the sample
        metadata attribute size.

        The region selection predicate can use all the classical equalities and disequalities
        {>, <, >=, <=, ==, !=} and predicates can be connected by the classical logical symbols
        {& (AND), | (OR), ~ (NOT)} plus the *isin* function.

        In order to be sure about the correctness of the expression, please use parenthesis to delimit
        the various predicates.
        """
        return self.select(region_predicate=predicate, semiJoinMeta=semiJoinMeta, semiJoinDataset=semiJoinDataset)

    def project(self, projected_meta=None, new_attr_dict=None, all_but_meta=None,
                projected_regs=None, new_field_dict=None, all_but_regs=None):
        projected_meta_exists = False
        if isinstance(projected_meta, list) and \
                all([isinstance(x, str) for x in projected_meta]):
            projected_meta = Some(projected_meta)
            projected_meta_exists = True
        elif projected_meta is None:
            projected_meta = none()
        else:
            raise TypeError("projected_meta must be a list of strings or None."
                            " {} was provided".format(type(projected_meta)))

        if isinstance(new_attr_dict, dict):
            meta_ext = []
            expBuild = self.pmg.getNewExpressionBuilder(self.__index)
            for k in new_attr_dict.keys():
                item = new_attr_dict[k]
                if isinstance(k, str):
                    if isinstance(item, MetaField):
                        me = expBuild.createMetaExtension(k, item.getMetaExpression())
                    elif isinstance(item, int):
                        me = expBuild.createMetaExtension(k, expBuild.getMEType("int", str(item)))
                    elif isinstance(item, str):
                        me = expBuild.createMetaExtension(k, expBuild.getMEType("string", item))
                    elif isinstance(item, float):
                        me = expBuild.createMetaExtension(k, expBuild.getMEType("float", str(item)))
                    else:
                        raise TypeError("Type {} of item of new_attr_dict is not valid".format(type(item)))
                    meta_ext.append(me)
                else:
                    raise TypeError("The key of new_attr_dict must be a string. "
                                    "{} was provided".format(type(k)))
            meta_ext = Some(meta_ext)
        elif new_attr_dict is None:
            meta_ext = none()
        else:
            raise TypeError("new_attr_dict must be a dictionary."
                            " {} was provided".format(type(new_attr_dict)))

        if isinstance(all_but_meta, list) and \
            all([isinstance(x, str) for x in all_but_meta]):
            if not projected_meta_exists:
                all_but_meta = Some(all_but_meta)
                all_but_value = True
            else:
                raise ValueError("all_but_meta and projected_meta are mutually exclusive")
        elif all_but_meta is None:
            all_but_meta = none()
            all_but_value = False
        else:
            raise TypeError("all_but_meta must be a list of strings."
                            " {} was provided".format(type(all_but_meta)))
        projected_meta = all_but_meta if all_but_value else projected_meta

        projected_regs_exists = False
        if isinstance(projected_regs, list) and \
                all([isinstance(x, str) for x in projected_regs]):
            projected_regs = Some(projected_regs)
            projected_regs_exists = True
        elif projected_regs is None:
            projected_regs = none()
        else:
            raise TypeError("projected_regs must be a list of strings or None."
                            " {} was provided".format(type(projected_regs)))

        if isinstance(new_field_dict, dict):
            regs_ext = []
            expBuild = self.pmg.getNewExpressionBuilder(self.__index)
            for k in new_field_dict.keys():
                item = new_field_dict[k]
                if isinstance(k, str):
                    if isinstance(item, RegField):
                        re = expBuild.createRegionExtension(k, item.getRegionExpression())
                    elif isinstance(item, MetaField):
                        re = expBuild.createRegionExtension(k, item.reMetaNode)
                    elif isinstance(item, int):
                        re = expBuild.createRegionExtension(k, expBuild.getREType("float", str(item)))
                    elif isinstance(item, str):
                        re = expBuild.createRegionExtension(k, expBuild.getREType("string", item))
                    elif isinstance(item, float):
                        re = expBuild.createRegionExtension(k, expBuild.getREType("float", str(item)))
                    else:
                        raise TypeError("Type {} of item of new_field_dict is not valid".format(type(item)))
                    regs_ext.append(re)
                else:
                    raise TypeError("The key of new_field_dict must be a string. "
                                    "{} was provided".format(type(k)))
            regs_ext = Some(regs_ext)
        elif new_field_dict is None:
            regs_ext = none()
        else:
            raise TypeError("new_field_dict must be a dictionary."
                            " {} was provided".format(type(new_field_dict)))

        if isinstance(all_but_regs, list) and \
            all([isinstance(x, str) for x in all_but_regs]):
            if not projected_regs_exists:
                all_but_regs = Some(all_but_regs)
            else:
                raise ValueError("all_but_meta and projected_meta are mutually exclusive")
        elif all_but_regs is None:
            all_but_regs = none()
        else:
            raise TypeError("all_but_regs must be a list of strings."
                            " {} was provided".format(type(all_but_regs)))

        new_index = self.opmng.project(self.__index, projected_meta, meta_ext, all_but_value,
                                       projected_regs, all_but_regs, regs_ext)
        return GMQLDataset(index=new_index, location=self.location,
                           local_sources=self._local_sources, remote_sources=self._remote_sources,
                           meta_profile=self.meta_profile)

    def meta_project(self, attr_list=None, all_but=None, new_attr_dict=None):
        """
        Project the metadata based on a list of attribute names
        
        :param attr_list: list of the metadata fields to select
        :param all_but: list of metadata that must be excluded from the projection.
        :param new_attr_dict: an optional dictionary of the form {'new_field_1': function1,
               'new_field_2': function2, ...} in which every function computes
               the new field based on the values of the others
        :return: a new GMQLDataset

        Notice that if attr_list is specified, all_but cannot be specified and viceversa.

        Examples::

            new_dataset = dataset.meta_project(attr_list=['antibody', 'ID'],
                                               new_attr_dict={'new_meta': dataset['ID'] + 100})

        """
        return self.project(projected_meta=attr_list, new_attr_dict=new_attr_dict, all_but_meta=all_but)

    def reg_project(self, field_list=None, all_but=None, new_field_dict=None):
        """
        Project the region data based on a list of field names

        :param field_list: list of the fields to select
        :param all_but: keep only the region fields different from the ones
               specified
        :param new_field_dict: an optional dictionary of the form {'new_field_1':
               function1, 'new_field_2': function2, ...} in which every function
               computes the new field based on the values of the others
        :return: a new GMQLDataset

        An example of usage::

            new_dataset = dataset.reg_project(['pValue', 'name'],
                                            {'new_field': dataset.pValue / 2})

            new_dataset = dataset.reg_project(field_list=['peak', 'pvalue'],
                                              new_field_dict={'new_field': dataset.pvalue * dataset['cell_age', 'float']})

        Notice that you can use metadata attributes for building new region fields.
        The only thing to remember when doing so is to define also the type of the output region field
        in the metadata attribute definition (for example, :code:`dataset['cell_age', 'float']` is
        required for defining the new attribute :code:`new_field` as float).
        In particular, the following type names are accepted: 'string', 'char', 'long', 'integer',
        'boolean', 'float', 'double'.

        """
        return self.project(projected_regs=field_list, all_but_regs=all_but, new_field_dict=new_field_dict)

    def extend(self, new_attr_dict):
        """
        For each sample in an input dataset, the EXTEND operator builds new metadata attributes,
        assigns their values as the result of arithmetic and/or aggregate functions calculated on 
        sample region attributes, and adds them to the existing metadata attribute-value pairs of 
        the sample. Sample number and their genomic regions, with their attributes and values, 
        remain unchanged in the output dataset.  
        
        :param new_attr_dict: a dictionary of the type {'new_metadata' : AGGREGATE_FUNCTION('field'), ...}
        :return: new GMQLDataset
        
        An example of usage::
        
            import gmql as gl
            
            # ....previous operations on the dataset
            
            new_dataset = dataset.extend({'regionCount' : gl.COUNT(),
                                          'minPValue' : gl.MIN('pValue')})
        """
        if isinstance(new_attr_dict, dict):
            expBuild = self.pmg.getNewExpressionBuilder(self.__index)
            aggregates = []
            for k in new_attr_dict.keys():
                if isinstance(k, str):
                    item = new_attr_dict[k]
                    if isinstance(item, Aggregate):
                        op_name = item.get_aggregate_name()
                        op_argument = Some(item.get_argument()) if item.is_unary() else none()
                        regsToMeta = expBuild.getRegionsToMeta(op_name, k, op_argument)
                        aggregates.append(regsToMeta)
                    else:
                        raise TypeError("The items in new_reg_fields must be Aggregates."
                                        " {} was provided".format(type(item)))
                else:
                    raise TypeError("The key of new_attr_dict must be a string. "
                                    "{} was provided".format(type(k)))
        else:
            raise TypeError("new_attr_dict must be a dictionary. "
                            "{} was provided".format(type(new_attr_dict)))

        new_index = self.opmng.extend(self.__index, aggregates)
        return GMQLDataset(index=new_index, location=self.location,
                           local_sources=self._local_sources,
                           remote_sources=self._remote_sources, meta_profile=self.meta_profile)

    def cover(self, minAcc, maxAcc, groupBy=None, new_reg_fields=None, cover_type="normal"):
        """
        COVER is a GMQL operator that takes as input a dataset (of usually, 
        but not necessarily, multiple samples) and returns another dataset 
        (with a single sample, if no groupby option is specified) by “collapsing”
        the input samples and their regions according to certain rules 
        specified by the COVER parameters. The attributes of the output regions 
        are only the region coordinates, plus in case, when aggregate functions 
        are specified, new attributes with aggregate values over attribute values 
        of the contributing input regions; output metadata are the union of the 
        input ones, plus the metadata attributes JaccardIntersect and 
        JaccardResult, representing global Jaccard Indexes for the considered 
        dataset, computed as the correspondent region Jaccard Indexes but on 
        the whole sample regions.

        :param cover_type: the kind of cover variant you want ['normal', 'flat', 'summit', 'histogram']
        :param minAcc: minimum accumulation value, i.e. the minimum number
         of overlapping regions to be considered during COVER execution. It can be any positive
         number or the strings {'ALL', 'ANY'}.
        :param maxAcc: maximum accumulation value, i.e. the maximum number
         of overlapping regions to be considered during COVER execution. It can be any positive
         number or the strings {'ALL', 'ANY'}.
        :param groupBy: optional list of metadata attributes 
        :param new_reg_fields: dictionary of the type
            {'new_region_attribute' : AGGREGATE_FUNCTION('field'), ...}
        :return: a new GMQLDataset
        
        An example of usage::
        
            cell_tf = narrow_peak.cover("normal", minAcc=1, maxAcc="Any", 
                                            groupBy=['cell', 'antibody_target'])    
        """
        if isinstance(cover_type, str):
            coverFlag = self.opmng.getCoverTypes(cover_type)
        else:
            raise TypeError("type must be a string. "
                            "{} was provided".format(type(cover_type)))
        if isinstance(minAcc, str):
            minAccParam = self.opmng.getCoverParam(minAcc.lower())
        elif isinstance(minAcc, int):
            minAccParam = self.opmng.getCoverParam(str(minAcc).lower())
        else:
            raise TypeError("minAcc must be a string or an integer. "
                            "{} was provided".format(type(minAcc)))
        if isinstance(maxAcc, str):
            maxAccParam = self.opmng.getCoverParam(maxAcc.lower())
        elif isinstance(maxAcc, int):
            maxAccParam = self.opmng.getCoverParam(str(maxAcc).lower())
        else:
            raise TypeError("maxAcc must be a string or an integer. "
                            "{} was provided".format(type(minAcc)))

        if isinstance(groupBy, list) and \
            all([isinstance(x, str) for x in groupBy]):
            groupBy_result = Some(groupBy)
        elif groupBy is None:
            groupBy_result = none()
        else:
            raise TypeError("groupBy must be a list of string. "
                            "{} was provided".format(type(groupBy)))

        aggregates = []
        if isinstance(new_reg_fields, dict):
            expBuild = self.pmg.getNewExpressionBuilder(self.__index)
            for k in new_reg_fields.keys():
                if isinstance(k, str):
                    item = new_reg_fields[k]
                    if isinstance(item, (SUM, MIN, MAX, AVG, BAG, BAGD,
                                         MEDIAN, COUNT)):
                        op_name = item.get_aggregate_name()
                        op_argument = item.get_argument()
                        if op_argument is None:
                            op_argument = none()
                        else:
                            op_argument = Some(op_argument)
                        regsToReg = expBuild.getRegionsToRegion(op_name, k, op_argument)
                        aggregates.append(regsToReg)
                    else:
                        raise TypeError("The items in new_reg_fields must be Aggregates (SUM, MIN, MAX, AVG, BAG, "
                                        "BAGD, MEDIAN, COUNT)"
                                        " {} was provided".format(type(item)))
                else:
                    raise TypeError("The key of new_reg_fields must be a string. "
                                    "{} was provided".format(type(k)))
        elif new_reg_fields is None:
            pass
        else:
            raise TypeError("new_reg_fields must be a list of dictionary. "
                            "{} was provided".format(type(new_reg_fields)))

        new_index = self.opmng.cover(self.__index, coverFlag, minAccParam, maxAccParam,
                                     groupBy_result, aggregates)
        return GMQLDataset(index=new_index, location=self.location,
                           local_sources=self._local_sources,
                           remote_sources=self._remote_sources, meta_profile=self.meta_profile)

    def normal_cover(self, minAcc, maxAcc, groupBy=None, new_reg_fields=None):
        """
        The normal cover operation as described in :meth:`~.cover`.
        Equivalent to calling::
        
            dataset.cover("normal", ...)
        """
        return self.cover(minAcc, maxAcc, groupBy, new_reg_fields, cover_type="normal")

    def flat_cover(self, minAcc, maxAcc, groupBy=None, new_reg_fields=None):
        """
        Variant of the function :meth:`~.cover` that returns the union of all the regions 
        which contribute to the COVER. More precisely, it returns the contiguous regions 
        that start from the first end and stop at the last end of the regions which would 
        contribute to each region of a COVER.
        
        Equivalent to calling::
        
            cover("flat", ...)
        """
        return self.cover(minAcc, maxAcc, groupBy, new_reg_fields, cover_type="flat")

    def summit_cover(self, minAcc, maxAcc, groupBy=None, new_reg_fields=None):
        """
        Variant of the function :meth:`~.cover` that returns only those portions of the COVER
        result where the maximum number of regions overlap (this is done by returning only
        regions that start from a position after which the number of overlaps does not 
        increase, and stop at a position where either the number of overlapping regions decreases 
        or violates the maximum accumulation index).
        
        Equivalent to calling::
        
            cover("summit", ...)
        """
        return self.cover(minAcc, maxAcc, groupBy, new_reg_fields, cover_type="summit")

    def histogram_cover(self, minAcc, maxAcc, groupBy=None, new_reg_fields=None):
        """
        Variant of the function :meth:`~.cover` that returns all regions contributing to 
        the COVER divided in different (contiguous) parts according to their accumulation 
        index value (one part for each different accumulation value), which is assigned to 
        the AccIndex region attribute.
        
        Equivalent to calling::
        
            cover("histogram", ...)
        """
        return self.cover(minAcc, maxAcc, groupBy, new_reg_fields, cover_type="histogram")

    def join(self, experiment, genometric_predicate, output="LEFT", joinBy=None,
             refName="REF", expName="EXP", left_on=None, right_on=None):
        """ The JOIN operator takes in input two datasets, respectively known as anchor
        (the first/left one) and experiment (the second/right one) and returns a
        dataset of samples consisting of regions extracted from the operands 
        according to the specified condition (known as genometric predicate). 
        The number of generated output samples is the Cartesian product of the number 
        of samples in the anchor and in the experiment dataset (if no joinby close 
        if specified). The attributes (and their values) of the regions in the output 
        dataset are the union of the region attributes (with their values) in the input 
        datasets; homonymous attributes are disambiguated by prefixing their name with 
        their dataset name. The output metadata are the union of the input metadata, 
        with their attribute names prefixed with their input dataset name.
        
        :param experiment: an other GMQLDataset
        :param genometric_predicate: a list of Genometric atomic conditions
        :param output: one of four different values that declare which region is given in output 
                       for each input pair of anchor and experiment regions satisfying the genometric predicate:
                       
                       * 'LEFT': outputs the anchor regions from the anchor dataset that satisfy the genometric predicate
                       * 'RIGHT': outputs the anchor regions from the experiment dataset that satisfy the genometric predicate
                       * 'INT': outputs the overlapping part (intersection) of the anchor and experiment regions that satisfy
                         the genometric predicate; if the intersection is empty, no output is produced
                       * 'CONTIG': outputs the concatenation between the anchor and experiment regions that satisfy the 
                         genometric predicate, i.e. the output region is defined as having left (right) coordinates 
                         equal to the minimum (maximum) of the corresponding coordinate values in the anchor and 
                         experiment regions satisfying the genometric predicate


        :param joinBy: list of metadata attributes
        :param refName: name that you want to assign to the reference dataset
        :param expName: name that you want to assign to the experiment dataset
        :param left_on: list of region fields of the reference on which the join must be performed
        :param right_on: list of region fields of the experiment on which the join must be performed
        :return: a new GMQLDataset
        
        An example of usage::
            
            import gmql as gl
            
            # anchor_dataset and experiment_dataset are created
            
            result_dataset = anchor_dataset.join(experiment=experiment_dataset, 
                                                    genometric_predicate=[gl.MD(1), gl.DGE(120000)], 
                                                    output="right")
        """

        if isinstance(experiment, GMQLDataset):
            other_idx = experiment.__index
        else:
            raise TypeError("experiment must be a GMQLDataset. "
                            "{} was provided".format(type(experiment)))
        if isinstance(genometric_predicate, list) and \
            all([isinstance(x, GenometricCondition) for x in genometric_predicate]):
            regionJoinCondition = self.opmng.getRegionJoinCondition(list(map(lambda x: x.get_gen_condition(),
                                                                             genometric_predicate)))
        else:
            raise TypeError("genometric_predicate must be a list og GenometricCondition. "
                            "{} was found".format(type(genometric_predicate)))

        if isinstance(output, str):
            regionBuilder = self.opmng.getRegionBuilderJoin(output)
        else:
            raise TypeError("output must be a string. "
                            "{} was provided".format(type(output)))

        if not isinstance(expName, str):
            raise TypeError("expName must be a string. {} was provided".format(type(expName)))

        if not isinstance(refName, str):
            raise TypeError("refName must be a string. {} was provided".format(type(expName)))

        if isinstance(joinBy, list) and \
            all([isinstance(x, str) for x in joinBy]):
            metaJoinCondition = Some(self.opmng.getMetaJoinCondition(joinBy))
        elif joinBy is None:
            metaJoinCondition = none()
        else:
            raise TypeError("joinBy must be a list of strings. "
                            "{} was found".format(type(joinBy)))

        left_on_exists = False
        left_on_len = 0
        if isinstance(left_on, list) and \
            all([isinstance(x, str) for x in left_on]):
            left_on_len = len(left_on)
            left_on = Some(left_on)
            left_on_exists = True
        elif left_on is None:
            left_on = none()
        else:
            raise TypeError("left_on must be a list of strings. "
                            "{} was provided".format(type(left_on)))

        if isinstance(right_on, list) and \
            all([isinstance(x, str)] for x in right_on) and \
            left_on_exists and len(right_on) == left_on_len:
            right_on = Some(right_on)
        elif right_on is None and not left_on_exists:
            right_on = none()
        else:
            raise TypeError("right_on must be a list of strings. "
                            "{} was provided".format(type(right_on)))

        new_index = self.opmng.join(self.__index, other_idx,
                                    metaJoinCondition, regionJoinCondition, regionBuilder,
                                    refName, expName, left_on, right_on)
        new_local_sources, new_remote_sources = self.__combine_sources(self, experiment)
        new_location = self.__combine_locations(self, experiment)
        return GMQLDataset(index=new_index, location=new_location,
                           local_sources=new_local_sources,
                           remote_sources=new_remote_sources, meta_profile=self.meta_profile)

    def map(self, experiment, new_reg_fields=None, joinBy=None, refName="REF", expName="EXP"):
        """
        MAP is a non-symmetric operator over two datasets, respectively called
        reference and experiment. The operation computes, for each sample in 
        the experiment dataset, aggregates over the values of the experiment 
        regions that intersect with a region in a reference sample, for each 
        region of each sample in the reference dataset; we say that experiment 
        regions are mapped to the reference regions.
        The number of generated output samples is the Cartesian product of the samples 
        in the two input datasets; each output sample has the same regions as the related 
        input reference sample, with their attributes and values, plus the attributes 
        computed as aggregates over experiment region values. Output sample metadata 
        are the union of the related input sample metadata, whose attribute names are 
        prefixed with their input dataset name. 
        For each reference sample, the MAP operation produces a matrix like structure,
        called genomic space, where each experiment sample is associated with a row, 
        each reference region with a column, and each matrix row is a vector of numbers
        - the aggregates computed during MAP execution. When the features of the reference 
        regions are unknown, the MAP helps in extracting the most interesting regions 
        out of many candidates. 
        
        :param experiment: a GMQLDataset
        :param new_reg_fields: an optional dictionary of the form 
               {'new_field_1': AGGREGATE_FUNCTION(field), ...}
                                
        :param joinBy: optional list of metadata
        :param refName: name that you want to assign to the reference dataset
        :param expName: name that you want to assign to the experiment dataset
        :return: a new GMQLDataset
        """
        if isinstance(experiment, GMQLDataset):
            other_idx = experiment.__index
        else:
            raise TypeError("experiment must be a GMQLDataset. "
                            "{} was provided".format(type(experiment)))

        aggregates = []
        if isinstance(new_reg_fields, dict):
            expBuild = self.pmg.getNewExpressionBuilder(experiment.__index)
            for k in new_reg_fields.keys():
                if isinstance(k, str):
                    item = new_reg_fields[k]
                    if isinstance(item, (SUM, MIN, MAX, AVG, BAG, BAGD,
                                         MEDIAN, COUNT)):
                        op_name = item.get_aggregate_name()
                        op_argument = item.get_argument()
                        if op_argument is None:
                            op_argument = none()
                        else:
                            op_argument = Some(op_argument)
                        regsToReg = expBuild.getRegionsToRegion(op_name, k, op_argument)
                        aggregates.append(regsToReg)
                    else:
                        raise TypeError("The items in new_reg_fields must be Aggregates (SUM, MIN, MAX, AVG, BAG, BAGD, "
                                        "MEDIAN, COUNT)"
                                        " {} was provided".format(type(item)))
                else:
                    raise TypeError("The key of new_reg_fields must be a string. "
                                    "{} was provided".format(type(k)))
        elif new_reg_fields is None:
            pass
        else:
            raise TypeError("new_reg_fields must be a list of dictionary. "
                            "{} was provided".format(type(new_reg_fields)))

        if isinstance(joinBy, list) and \
            all([isinstance(x, str) for x in joinBy]):
            metaJoinCondition = Some(self.opmng.getMetaJoinCondition(joinBy))
        elif joinBy is None:
            metaJoinCondition = none()
        else:
            raise TypeError("joinBy must be a list of strings. "
                            "{} was found".format(type(joinBy)))

        if not isinstance(expName, str):
            raise TypeError("expName must be a string. {} was provided".format(type(expName)))

        if not isinstance(refName, str):
            raise TypeError("refName must be a string. {} was provided".format(type(expName)))

        new_index = self.opmng.map(self.__index, other_idx, metaJoinCondition,
                                   aggregates, refName, expName)
        new_local_sources, new_remote_sources = self.__combine_sources(self, experiment)
        new_location = self.__combine_locations(self, experiment)
        return GMQLDataset(index=new_index, location=new_location,
                           local_sources=new_local_sources,
                           remote_sources=new_remote_sources,
                           meta_profile=self.meta_profile)

    def order(self, meta=None, meta_ascending=None, meta_top=None, meta_k=None,
              regs=None, regs_ascending=None, region_top=None, region_k=None):
        """ The ORDER operator is used to order either samples, sample regions, or both,
        in a dataset according to a set of metadata and/or region attributes, and/or 
        region coordinates. The number of samples and their regions in the output dataset 
        is as in the input dataset, as well as their metadata and region attributes 
        and values, but a new ordering metadata and/or region attribute is added with 
        the sample or region ordering value, respectively. 
        
        :param meta: list of metadata attributes
        :param meta_ascending: list of boolean values (True = ascending, False = descending)
        :param meta_top: "top", "topq" or "topp" or None
        :param meta_k: a number specifying how many results to be retained
        :param regs: list of region attributes
        :param regs_ascending: list of boolean values (True = ascending, False = descending)
        :param region_top: "top", "topq" or "topp" or None
        :param region_k: a number specifying how many results to be retained
        :return: a new GMQLDataset
        """
        meta_exists = False
        meta_len = 0
        if isinstance(meta, list) and \
            all([isinstance(x, str) for x in meta]):
            meta_exists = True
            meta_len = len(meta)
            meta = Some(meta)
        elif meta is None:
            meta = none()
        else:
            raise TypeError("meta must be a list of strings. "
                            "{} was provided".format(type(meta)))

        if isinstance(meta_ascending, list) and \
            all([isinstance(x, bool) for x in meta_ascending]) and \
            meta_exists and meta_len == len(meta_ascending):
            meta_ascending = Some(meta_ascending)
        elif meta_ascending is None:
            if meta_exists:
                # by default meta_ascending is all True
                meta_ascending = Some([True for _ in range(meta_len)])
            else:
                meta_ascending = none()
        else:
            raise TypeError("meta ascending must be a list of booleans having the same size "
                            "of meta. {} was provided".format(type(meta_ascending)))

        regs_exists = False
        regs_len = 0
        if isinstance(regs, list) and \
                all([isinstance(x, str) for x in regs]):
            regs_exists = True
            regs_len = len(regs)
            regs = Some(regs)
        elif regs is None:
            regs = none()
        else:
            raise TypeError("regs must be a list of strings. "
                            "{} was provided".format(type(regs)))

        if isinstance(regs_ascending, list) and \
            all([isinstance(x, bool) for x in regs_ascending]) and \
                regs_exists and regs_len == len(regs_ascending):
            regs_ascending = Some(regs_ascending)
        elif regs_ascending is None:
            if regs_exists:
                # by default regs_ascending is all True
                regs_ascending = Some([True for _ in range(regs_len)])
            else:
                regs_ascending = none()
        else:
            raise TypeError("meta regs_ascending must be a list of booleans having the same size "
                            "of regs. {} was provided".format(type(regs_ascending)))

        meta_top_exists = False
        if isinstance(meta_top, str):
            if meta_exists:
                meta_top = Some(meta_top)
                meta_top_exists = True
            else:
                raise ValueError("meta_top must be defined only when meta is defined")
        elif meta_top is None:
            meta_top = none()
        else:
            raise TypeError("meta_top must be a string. {} was provided".format(type(meta_top)))

        if isinstance(meta_k, int) and \
                meta_top_exists:
            meta_k = Some(meta_k)
        elif meta_k is None and \
                not meta_top_exists:
            meta_k = none()
        else:
            raise TypeError("meta_k must be an integer and should be provided together with a "
                            "value of meta_top. {} was provided".format(type(meta_k)))

        region_top_exists = False
        if isinstance(region_top, str):
            if regs_exists:
                region_top = Some(region_top)
                region_top_exists = True
            else:
                raise ValueError("region_top must be defined only when regs is defined")
        elif region_top is None:
            region_top = none()
        else:
            raise TypeError("region_top must be a string. {} was provided".format(type(region_top)))

        if isinstance(region_k, int) and \
                region_top_exists:
            region_k = Some(region_k)
        elif region_k is None and \
                not region_top_exists:
            region_k = none()
        else:
            raise TypeError("region_k must be an integer and should be provided together with a "
                            "value of region_top. {} was provided".format(type(region_k)))

        new_index = self.opmng.order(self.__index, meta, meta_ascending, meta_top, meta_k,
                                     regs, regs_ascending, region_top, region_k)
        return GMQLDataset(index=new_index, location=self.location,
                           local_sources=self._local_sources,
                           remote_sources=self._remote_sources,
                           meta_profile=self.meta_profile)

    def difference(self, other, joinBy=None, exact=False):
        """ DIFFERENCE is a binary, non-symmetric operator that produces one sample
        in the result for each sample of the first operand, by keeping the same
        metadata of the first operand sample and only those regions (with their
        schema and values) of the first operand sample which do not intersect with
        any region in the second operand sample (also known as negative regions)
        
        :param other: GMQLDataset
        :param joinBy: list of metadata attributes
        :param exact: boolean
        :return: a new GMQLDataset
        """

        if isinstance(other, GMQLDataset):
            other_idx = other.__index
        else:
            raise TypeError("other must be a GMQLDataset. "
                            "{} was provided".format(type(other)))

        if isinstance(joinBy, list) and \
            all([isinstance(x, str) for x in joinBy]):
            metaJoinCondition = Some(self.opmng.getMetaJoinCondition(joinBy))
        elif joinBy is None:
            metaJoinCondition = none()
        else:
            raise TypeError("joinBy must be a list of strings. "
                            "{} was provided".format(type(joinBy)))

        if not isinstance(exact, bool):
            raise TypeError("exact must be a boolean. "
                            "{} was provided".format(type(exact)))

        new_index = self.opmng.difference(self.__index, other_idx, metaJoinCondition, exact)

        new_local_sources, new_remote_sources = self.__combine_sources(self, other)
        new_location = self.__combine_locations(self, other)
        return GMQLDataset(index=new_index, location=new_location,
                           local_sources=new_local_sources,
                           remote_sources=new_remote_sources,
                           meta_profile=self.meta_profile)

    def union(self, other, left_name="LEFT", right_name="RIGHT"):
        """
        The UNION operation is used to integrate homogeneous or heterogeneous samples of two
        datasets within a single dataset; for each sample of either one of the input datasets, a
        sample is created in the result as follows:
        
            * its metadata are the same as in the original sample;
            * its schema is the schema of the first (left) input dataset; new 
              identifiers are assigned to each output sample;
            * its regions are the same (in coordinates and attribute values) as in the original
              sample. Region attributes which are missing in an input dataset sample
              (w.r.t. the merged schema) are set to null.
                
        :param other: a GMQLDataset
        :param left_name: name that you want to assign to the left dataset
        :param right_name: name that you want to assign to the right dataset
        :return: a new GMQLDataset
        """

        if not isinstance(left_name, str) or \
            not isinstance(right_name, str):
            raise TypeError("left_name and right_name must be strings. "
                            "{} - {} was provided".format(type(left_name), type(right_name)))

        if isinstance(other, GMQLDataset):
            other_idx = other.__index
        else:
            raise TypeError("other must be a GMQLDataset. "
                            "{} was provided".format(type(other)))

        if len(left_name) == 0 or len(right_name) == 0:
            raise ValueError("left_name and right_name must not be empty")
        new_index = self.opmng.union(self.__index, other_idx, left_name, right_name)
        new_local_sources, new_remote_sources = self.__combine_sources(self, other)
        new_location = self.__combine_locations(self, other)
        return GMQLDataset(index=new_index, location=new_location,
                           local_sources=new_local_sources,
                           remote_sources=new_remote_sources,
                           meta_profile=self.meta_profile)

    def merge(self, groupBy=None):
        """
        The MERGE operator builds a new dataset consisting of a single sample having
        
            * as regions all the regions of all the input samples, with the 
              same attributes and values
            * as metadata the union of all the metadata attribute-values 
              of the input samples.
                
        A groupby clause can be specified on metadata: the samples are then
        partitioned in groups, each with a distinct value of the grouping metadata 
        attributes, and the MERGE operation is applied to each group separately, 
        yielding to one sample in the result dataset for each group.
        Samples without the grouping metadata attributes are disregarded
        
        :param groupBy: list of metadata attributes
        :return: a new GMQLDataset
        """

        if isinstance(groupBy, list) and \
            all([isinstance(x, str) for x in groupBy]):
            groupBy = Some(groupBy)
        elif groupBy is None:
            groupBy = none()
        else:
            raise TypeError("groupBy must be a list of strings. "
                            "{} was provided".format(type(groupBy)))

        new_index = self.opmng.merge(self.__index, groupBy)
        return GMQLDataset(index=new_index, location=self.location,
                           local_sources=self._local_sources,
                           remote_sources=self._remote_sources,
                           meta_profile=self.meta_profile)

    def group(self, meta=None, meta_aggregates=None, regs=None,
              regs_aggregates=None, meta_group_name="_group"):
        """ The GROUP operator is used for grouping both regions and/or metadata of input
        dataset samples according to distinct values of certain attributes (known as grouping
        attributes); new grouping attributes are added to samples in the output dataset,
        storing the results of aggregate function evaluations over metadata and/or regions
        in each group of samples.
        Samples having missing values for any of the grouping attributes are discarded.


        :param meta: (optional) a list of metadata attributes
        :param meta_aggregates: (optional) {'new_attr': fun}
        :param regs: (optional) a list of region fields
        :param regs_aggregates: {'new_attr': fun}
        :param meta_group_name: (optional) the name to give to the group attribute in the
               metadata
        :return: a new GMQLDataset
        """

        if isinstance(meta, list) and \
            all([isinstance(x, str) for x in meta]):
            meta = Some(meta)
        elif meta is None:
            meta = none()
        else:
            raise TypeError("meta must be a list of strings. "
                            "{} was provided".format(type(meta)))
        expBuild = self.pmg.getNewExpressionBuilder(self.__index)
        if isinstance(meta_aggregates, dict):
            metaAggregates = []
            for k in meta_aggregates:
                if isinstance(k, str):
                    item = meta_aggregates[k]
                    if isinstance(item, (SUM, MIN, MAX, AVG, BAG,
                                         BAGD, STD, MEDIAN, COUNTSAMP)):
                        functionName = item.get_aggregate_name()
                        argument = item.get_argument()
                        if argument is None:
                            argument = none()
                        else:
                            argument = Some(argument)
                        metaAggregates.append(expBuild.createMetaAggregateFunction(functionName,
                                                                                   k, argument))
                    else:
                        raise TypeError("the item of the dictionary must be an Aggregate of the following: "
                                        "SUM, MIN, MAX, AVG, BAG, BAGD, STD, COUNTSAMP. "
                                        "{} was provided".format(type(item)))
                else:
                    raise TypeError("keys of meta_aggregates must be string. "
                                    "{} was provided".format(type(k)))
            metaAggregates = Some(metaAggregates)
        elif meta_aggregates is None:
            metaAggregates = none()
        else:
            raise TypeError("meta_aggregates must be a dictionary of Aggregate functions. "
                            "{} was provided".format(type(meta_aggregates)))
        if isinstance(regs, list) and \
            all([isinstance(x, str) for x in regs]):
            regs = Some(regs)
        elif regs is None:
            regs = none()
        else:
            raise TypeError("regs must be a list of strings. "
                            "{} was provided".format(type(regs)))

        if isinstance(regs_aggregates, dict):
            regionAggregates = []
            for k in regs_aggregates.keys():
                if isinstance(k, str):
                    item = regs_aggregates[k]
                    if isinstance(item, (SUM, MIN, MAX, AVG, BAG, BAGD,
                                         MEDIAN, COUNT)):
                        op_name = item.get_aggregate_name()
                        op_argument = item.get_argument()
                        if op_argument is None:
                            op_argument = none()
                        else:
                            op_argument = Some(op_argument)
                        regsToReg = expBuild.getRegionsToRegion(op_name, k, op_argument)
                        regionAggregates.append(regsToReg)
                    else:
                        raise TypeError("the item of the dictionary must be an Aggregate of the following: "
                                        "SUM, MIN, MAX, AVG, BAG, BAGD, MEDIAN, COUNT. "
                                        "{} was provided".format(type(item)))
                else:
                    raise TypeError("The key of new_reg_fields must be a string. "
                                    "{} was provided".format(type(k)))
            regionAggregates = Some(regionAggregates)
        elif regs_aggregates is None:
            regionAggregates = none()
        else:
            raise TypeError("new_reg_fields must be a list of dictionary. "
                            "{} was provided".format(type(regs_aggregates)))

        if isinstance(meta_group_name, str):
            pass
        else:
            raise TypeError("meta_group_name must be a string. "
                            "{} was provided".format(type(meta_group_name)))

        new_index = self.opmng.group(self.__index, meta, metaAggregates, meta_group_name, regs, regionAggregates)

        return GMQLDataset(index=new_index, location=self.location,
                           local_sources=self._local_sources,
                           remote_sources=self._remote_sources,
                           meta_profile=self.meta_profile)

    def meta_group(self, meta, meta_aggregates=None):
        """ Group operation only for metadata. For further information check :meth:`~.group`
        """
        return self.group(meta=meta, meta_aggregates=meta_aggregates)

    def regs_group(self, regs, regs_aggregates=None):
        """ Group operation only for region data. For further information check :meth:`~.group`
        """
        return self.group(regs=regs, regs_aggregates=regs_aggregates)
    """
        Materialization utilities
    """
    def materialize(self, output_path=None, output_name=None,  all_load=True):
        """ Starts the execution of the operations for the GMQLDataset. PyGMQL implements lazy execution
        and no operation is performed until the materialization of the results is requestd.
        This operation can happen both locally or remotely.

        * Local mode: if the GMQLDataset is local (based on local data) the user can specify the

        :param output_path: (Optional) If specified, the user can say where to locally save the results
                            of the computations.
        :param output_name: (Optional) Can be used only if the dataset is remote. It represents the name that
                            the user wants to give to the resulting dataset on the server
        :param all_load: (Optional) It affects the computation only when the dataset is remote. It specifies if
                         the downloaded result dataset should be directly converted to a GDataframe (True) or to a
                         GMQLDataset (False) for future local queries.
        :return: A GDataframe or a GMQLDataset
        """
        current_mode = get_mode()
        new_index = self.__modify_dag(current_mode)
        if current_mode == 'local':
            return Materializations.materialize_local(new_index, output_path)
        elif current_mode == 'remote':
            return Materializations.materialize_remote(new_index, output_name, output_path, all_load)
        else:
            raise ValueError("Current mode is not defined. {} given".format(current_mode))

    def take(self, n=5):
        """ Returns a small set of regions and metadata from a query. It is supposed to
        be used for debugging purposes or for data exploration.

        :param n: how many samples to retrieve
        :return: a GDataframe
        """
        if n <= 0:
            raise ValueError("n must be a positive number. {} was given".format(n))

        collected = self.pmg.take(self.__index, n)
        regs = MemoryLoader.load_regions(collected)
        meta = MemoryLoader.load_metadata(collected)
        result = GDataframe.GDataframe(regs, meta)
        return result

    def _get_serialized_dag(self):
        serialized_dag = self.pmg.serializeVariable(self.__index)
        return serialized_dag

    @staticmethod
    def __combine_sources(d1, d2):
        if (not isinstance(d1, GMQLDataset)) or (not isinstance(d2, GMQLDataset)):
            raise TypeError("The function takes only GMQLDataset")
        local_sources_1 = d1._local_sources
        remote_sources_1 = d1._remote_sources
        local_sources_2 = d2._local_sources
        remote_sources_2 = d2._remote_sources

        new_local_sources = list(set(local_sources_1 + local_sources_2))
        new_remote_sources = list(set(remote_sources_1 + remote_sources_2))

        return new_local_sources, new_remote_sources

    @staticmethod
    def __combine_locations(d1, d2):
        if (not isinstance(d1, GMQLDataset)) or (not isinstance(d2, GMQLDataset)):
            raise TypeError("The function takes only GMQLDataset")

        location_1 = d1.location
        location_2 = d2.location

        if location_1 == location_2:
            return location_1
        else:
            return "mixed"

    def __modify_dag(self, mode):
        remote_manager = get_remote_manager()
        index = self.__index
        pmg = get_python_manager()
        sources = _get_source_table()
        # create a new id having the exact same DAG inside, for modification
        new_index = pmg.cloneVariable(index)
        if mode == "local":
            for d in self._remote_sources:
                # for each remote source, we have to download it locally in a temporary folder
                d_sources = sources.get_source(id=d)
                local = d_sources[LOCAL]
                remote = d_sources[REMOTE]
                if local is None:
                    new_name = get_new_dataset_tmp_folder()
                    remote_manager.download_dataset(dataset_name=remote, local_path=new_name, how="stream")
                    sources.modify_source(id=d, local=new_name)
                else:
                    new_name = local
                pmg.modify_dag_source(new_index, str(d), new_name)
            for d in self._local_sources:
                # for each local source, just take its path
                d_sources = sources.get_source(id=d)
                local = d_sources[LOCAL]
                if local is None:
                    raise ValueError("Impossible state. Local source must have a local path")
                else:
                    pmg.modify_dag_source(new_index, str(d), local)
        elif mode == "remote":
            for d in self._local_sources:
                # for each local source, we have to upload it remotely
                d_sources = sources.get_source(id=d)
                local = d_sources[LOCAL]
                remote = d_sources[REMOTE]
                parser = d_sources[PARSER]
                if remote is None:
                    new_name = "LOCAL_" + get_unique_identifier()
                    schema_dir = get_new_dataset_tmp_folder()
                    os.makedirs(schema_dir)
                    schema_tmp_path = os.path.join(schema_dir, new_name + ".schema")
                    parserToXML(parser, new_name, schema_tmp_path)
                    remote_manager.upload_dataset(dataset=local, dataset_name=new_name, schema_path=schema_tmp_path)
                    sources.modify_source(id=d, remote=new_name)
                else:
                    new_name = remote
                pmg.modify_dag_source(new_index, str(d), new_name)
            for d in self._remote_sources:
                d_sources = sources.get_source(id=d)
                remote = d_sources[REMOTE]
                if remote is None:
                    raise ValueError("Impossible state. Remote source must have a remote name")
                else:
                    pmg.modify_dag_source(new_index, str(d), remote)
        else:
            raise ValueError("Unknown mode {}".format(mode))

        return new_index



from ... import get_python_manager
from . import RegField, _get_opposite_operator


class MetaField:
    def __init__(self, name, index=None, meta_condition=None, meNode=None, t=None, reMetaNode=None):
        if (meta_condition is not None) and (meNode is not None):
            raise ValueError("you cannot mix conditions and expressions")

        if isinstance(name, str):
            self.name = name
        else:
            raise TypeError("name of the metadata attribute must be a string")
        self.index = index
        pymg = get_python_manager()
        self.exp_build = pymg.getNewExpressionBuilder(self.index)

        if t is not None:
            if isinstance(t, str):
                self.t = t
                self.reMetaNode = self.exp_build.getREMetaAccessor(self.name, pymg.getParseTypeFromString(t))
            else:
                raise TypeError("Type of MetaField must be string")
        else:
            self.t = t
            self.reMetaNode = reMetaNode

        self.metaCondition = meta_condition

        if not (name.startswith("(") and name.endswith(")")) and meNode is None:
            self.meNode = self.exp_build.getMENode(name)
            self.metaAccessor = self.exp_build.getMetaAccessor(name)
        else:
            self.meNode = meNode

    def getMetaCondition(self):
        return self.metaCondition

    def getMetaExpression(self):
        return self.meNode

    """
        PREDICATES
    """

    def __eq__(self, other):
        return self._predicate(other, "EQ")

    def __ne__(self, other):
        return self._predicate(other, "NOTEQ")

    def __gt__(self, other):
        return self._predicate(other, "GT")

    def __ge__(self, other):
        return self._predicate(other, "GTE")

    def __lt__(self, other):
        return self._predicate(other, "LT")

    def __le__(self, other):
        return self._predicate(other, "LTE")

    def _predicate(self, other, operator):
        if isinstance(other, (int, str, float)):
            new_name = '(' + self.name + operator + str(other) + ')'
            predicate = self.exp_build.createMetaPredicate(self.name, operator, str(other))
            return MetaField(name=new_name, meta_condition=predicate, index=self.index)
        elif isinstance(other, RegField.RegField):
            return other._predicate(self, _get_opposite_operator(operator))
        else:
            raise TypeError("You can create predicates only using primitive types")

    def isin(self, values):
        """ Selects the samples having the metadata attribute between the values provided
        as input

        :param values: a list of elements
        :return a new complex condition
        """
        if not isinstance(values, list):
            raise TypeError("Input should be a string. {} was provided".format(type(values)))
        if not (self.name.startswith("(") and self.name.endswith(")")):
            first = True
            new_condition = None
            for v in values:
                if first:
                    first = False
                    new_condition = self.__eq__(v)
                else:
                    new_condition = new_condition.__or__(self.__eq__(v))
            return new_condition
        else:
            raise SyntaxError("You cannot use 'isin' with a complex condition")

    """
        CONDITIONS
    """

    def __and__(self, other):
        return self._binary_condition(other, "AND")

    def __or__(self, other):
        return self._binary_condition(other, "OR")

    def __invert__(self):
        return self._unary_condition("NOT")

    def _binary_condition(self, other, operator):
        if isinstance(other, MetaField):
            if other.index != self.index:
                raise ValueError("You cannot create conditions using different datasets")
            if other.metaCondition is None or \
                    self.metaCondition is None:
                raise ValueError("One of the two sides of the {} operator"
                                 " is not a condition".format(operator))
            new_name = '(' + self.name + operator + other.name + ')'
            condition = self.exp_build.createMetaBinaryCondition(self.metaCondition, operator, other.metaCondition)
            return MetaField(name=new_name, meta_condition=condition, index=self.index)
        else:
            raise TypeError("Binary conditions can be done only between MetaFields")

    def _unary_condition(self, operator):
        if self.metaCondition is None:
            raise ValueError("The inner part of the {} operator is not"
                             " a condition".format(operator))
        new_name = operator + '(' + self.name + ')'
        condition = self.exp_build.createMetaUnaryCondition(self.metaCondition, operator)
        return MetaField(name=new_name, meta_condition=condition, index=self.index)

    """
        EXPRESSIONS
    """

    def __add__(self, other):
        return self._binary_expression(other, "ADD")

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self._binary_expression(other, "SUB", order="left")

    def __rsub__(self, other):
        return self._binary_expression(other, "SUB", order="right")

    def __mul__(self, other):
        return self._binary_expression(other, "MUL")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._binary_expression(other, "DIV", order="left")

    def __rtruediv__(self, other):
        return self._binary_expression(other, "DIV", order="right")

    def __neg__(self):
        return self._unary_expression("NEG")

    def _binary_expression(self, other, operation, order="left"):
        if self.meNode is None:
            raise ValueError("Cannot mix expressions and conditions")
        if isinstance(other, RegField.RegField):
            return other._binary_expression(self, operation, "right" if order == "left" else "left")
        other, other_name, other_reNode = self.__get_return_type(other)
        renode = None
        if order == 'left':
            new_name = '(' + self.name + operation + other_name + ')'
            node = self.exp_build.getBinaryMetaExpression(self.meNode, operation, other)
            if other_reNode is not None:
                renode = self.exp_build.getBinaryRegionExpression(self.reMetaNode, operation, other_reNode)
        else:
            new_name = '(' + other_name + operation + self.name + ')'
            node = self.exp_build.getBinaryMetaExpression(other, operation, self.meNode)
            if other_reNode is not None:
                renode = self.exp_build.getBinaryRegionExpression(other_reNode, operation, self.reMetaNode)
        return MetaField(name=new_name, index=self.index, meNode=node, reMetaNode=renode)

    def _unary_expression(self, operation):
        if self.meNode is None:
            raise ValueError("Cannot mix expressions and conditions")
        new_name = operation + "(" + self.name + ")"
        node = self.exp_build.getUnaryMetaExpression(self.meNode, operation)
        renode = self.exp_build.getUnaryRegionExpression(self.reMetaNode, operation)
        return MetaField(name=new_name, index=self.index, meNode=node, reMetaNode=renode)

    def __get_return_type(self, other):
        if isinstance(other, str):
            other_reNode = self.exp_build.getREType("string", other)
            other_name = other
            other = self.exp_build.getMEType("string", other)
        elif isinstance(other, int):
            other_reNode = self.exp_build.getREType("int", str(other))
            other_name = str(other)
            other = self.exp_build.getMEType("int", str(other))
        elif isinstance(other, float):
            other_reNode = self.exp_build.getREType("float", str(other))
            other_name = str(other)
            other = self.exp_build.getMEType("float", str(other))
        elif isinstance(other, MetaField):
            other_name = other.name
            if other.index != self.index:
                raise ValueError("You cannot mix fields or attributes of different"
                                 "dataset in a condition or expression")
            if other.meNode is None:
                raise ValueError("You cannot mix conditions and expressions")
            other_reNode = other.reMetaNode
            other = other.meNode
        else:
            raise ValueError("Expected string, float, integer or MetaField. {} was found".format(type(other)))
        return other, other_name, other_reNode



from ... import get_python_manager, get_remote_manager, get_mode, _get_source_table
from ...FileManagment import TempFileManager
from ..parsers.RegionParser import RegionParser
from . import MetaLoaderFile, RegLoaderFile
from .. import GDataframe
from .. import GMQLDataset
import os
from .MetadataProfiler import create_metadata_profile


def get_file_paths(path):
    real_path = preprocess_path(path)
    all_files = os.listdir(real_path)

    def filter_files(x):
        return not (x.endswith(".xml") or x.endswith(".schema") \
                    or x.startswith("_"))

    files_paths = list(map(lambda x: os.path.join(real_path, x), filter(filter_files, all_files)))
    # files_paths = set(glob.glob(real_path + "/[!_]*"))
    schema_path = RegLoaderFile.get_schema_path(real_path)
    return files_paths, schema_path


def preprocess_path(path):
    for root, dirs, files in os.walk(path):
        if check_for_dataset(files):
            return root
    raise ValueError("The provided path does not contain any GMQL dataset")


def check_for_dataset(files):
    return len([x for x in files if x.endswith(".meta")]) > 0


def load_from_path(local_path=None, parser=None,  all_load=False):
    """ Loads the data from a local path into a GMQLDataset.
    The loading of the files is "lazy", which means that the files are loaded only when the
    user does a materialization (see :func:`~gmql.dataset.GMQLDataset.GMQLDataset.materialize` ).
    The user can force the materialization of the data (maybe for an initial data exploration on
    only the metadata) by setting the :attr:`~.reg_load` (load in memory the region data),
    :attr:`~.meta_load` (load in memory the metadata) or :attr:`~.all_load` (load both region and
    meta data in memory). If the user specifies this final parameter as True, a
    :class:`~gmql.dataset.GDataframe.GDataframe` is returned, otherwise a
    :class:`~gmql.dataset.GMQLDataset.GMQLDataset` is returned
    
    :param local_path: local path of the dataset
    :param parser: the parser to be used for reading the data
    :param all_load: if set to True, both region and meta data are loaded in memory and an 
                     instance of GDataframe is returned
    :return: A new GMQLDataset or a GDataframe
    """
    pmg = get_python_manager()

    local_path = preprocess_path(local_path)

    if all_load:
        # load directly the metadata for exploration
        meta = MetaLoaderFile.load_meta_from_path(local_path)
        if isinstance(parser, RegionParser):
            # region data
            regs = RegLoaderFile.load_reg_from_path(local_path, parser)
        else:
            regs = RegLoaderFile.load_reg_from_path(local_path)

        return GDataframe.GDataframe(regs=regs, meta=meta)
    else:
        from ... import _metadata_profiling
        if _metadata_profiling:
            meta_profile = create_metadata_profile(local_path)
        else:
            meta_profile = None

        index = None
        if parser is None:
            # find the parser
            parser = RegLoaderFile.get_parser(local_path)
        elif not isinstance(parser, RegionParser):
            raise ValueError("parser must be RegionParser. {} was provided".format(type(parser)))

        source_table = _get_source_table()
        id = source_table.search_source(local=local_path)
        if id is None:
            id = source_table.add_source(local=local_path, parser=parser)
        local_sources = [id]

        index = pmg.read_dataset(str(id), parser.get_gmql_parser())
        return GMQLDataset.GMQLDataset(index=index, parser=parser,
                                       location="local", path_or_name=local_path,
                                       local_sources=local_sources,
                                       meta_profile=meta_profile)


def load_from_remote(remote_name, owner=None):
    """ Loads the data from a remote repository.

    :param remote_name: The name of the dataset in the remote repository
    :param owner: (optional) The owner of the dataset. If nothing is provided, the current user
                  is used. For public datasets use 'public'.
    :return A new GMQLDataset or a GDataframe
    """
    pmg = get_python_manager()
    remote_manager = get_remote_manager()
    parser = remote_manager.get_dataset_schema(remote_name, owner)

    source_table = _get_source_table()
    id = source_table.search_source(remote=remote_name)
    if id is None:
        id = source_table.add_source(remote=remote_name, parser=parser)
    index = pmg.read_dataset(str(id), parser.get_gmql_parser())
    remote_sources = [id]
    return GMQLDataset.GMQLDataset(index=index, location="remote", path_or_name=remote_name,
                                   remote_sources=remote_sources)


def load(path=None, name=None, owner=None, parser=None, all_load=False):
    # TODO: think if this method is useful or not...
    mode = get_mode()
    remote_manager = get_remote_manager()
    if mode == 'local':
        if isinstance(path, str) and (name is None):
            # we are given a local path
            return load_from_path(local_path=path, parser=parser, all_load=all_load)
        elif isinstance(name, str) and (path is None):
            local_path = TempFileManager.get_new_dataset_tmp_folder()
            remote_manager.download_dataset(dataset_name=name, local_path=local_path)
            return load_from_path(local_path=local_path, all_load=all_load)
        else:
            ValueError("You have to define path or name (mutually exclusive)")
    elif mode == 'remote':
        if isinstance(path, str) and (name is None):
            name = TempFileManager.get_unique_identifier()
            remote_manager.upload_dataset(dataset=path, dataset_name=name)
            return load_from_remote(remote_name=name)
        elif isinstance(name, str) and (path is None):
            return load_from_remote(remote_name=name, owner=owner)
        else:
            ValueError("You have to define path or name (mutually exclusive)")
    else:
        raise ValueError("Mode: {} unknown".format(mode))




from glob import glob
import os
from tqdm import tqdm
import strconv
import pandas as pd


class MetadataProfile:

    def __init__(self, metadata_info):
        if isinstance(metadata_info, dict):
            self.metadata_info = metadata_info
        else:
            raise TypeError("metadata_info must be a dictionary. "
                            "{} was provided".format(type(metadata_info)))

    def get_metadata_type(self, metadata_attribute):
        if self.exists(metadata_attribute):
            t = self.metadata_info[metadata_attribute]
            return t
        else:
            raise ValueError("Metadata attribute {} not present".format(metadata_attribute))

    def get_metadata(self):
        return list(self.metadata_info.keys())

    def exists(self, metadata_attribute):
        return metadata_attribute in self.metadata_info.keys()

    def remove_attributes(self, l):
        if isinstance(l, str):
            if self.exists(l):
                self.metadata_info.pop(l)
            else:
                raise ValueError("Attribute {} doesn't exists".format(l))
        elif isinstance(l, list):
            for a in l:
                if self.exists(a):
                    self.metadata_info.pop(a)
                else:
                    raise ValueError("Attribute {} doesn't exists".format(a))
        else:
            raise TypeError("Input must be a string or a list")

    def select_attributes(self, l):
        if isinstance(l, list):
            res = dict()
            for a in l:
                if isinstance(a, str):
                    if self.exists(a):
                        res[a] = self.metadata_info[a]
                    else:
                        raise ValueError("Attribute {} doesn't exists".format(a))
                else:
                    raise TypeError("Attribute names must be strings")
        else:
            raise TypeError("Input must be a list")
        self.metadata_info = res

    def add_metadata(self, d):
        self.metadata_info.update(d)

    def to_df(self):
        df = pd.DataFrame.from_dict(self.metadata_info, orient="index")
        df.columns = ['Type', 'Values']
        # df['Type'] = df['Type'].map(str)
        df = df.sort_index()
        return df

    def show(self):
        print(self.to_df())

    def _repr_html_(self):
        return self.to_df().to_html()


def create_metadata_profile(dataset_path):
    meta_files = glob(pathname=dataset_path + '/*.meta')
    from ... import __disable_progress

    profile = dict()

    for mf in tqdm(meta_files, disable=__disable_progress):
        # print(mf + "\n\n")
        full_mf_path = os.path.abspath(mf)
        fo = open(full_mf_path)
        lines = fo.readlines()
        for l in lines:
            analyze_line(l, profile)
    return MetadataProfile(metadata_info=profile)


def analyze_line(line, d):
    fields = line.split("\t")
    name = fields[0]
    value = fields[1].strip()
    value_type = strconv.infer(value, converted=True)
    if value_type not in [str, int, float]:
        value_type = str
    # print("{} - {}".format(name, value))
    if name not in d.keys():
        d[name] = (value_type, set([value_type(value)]))
    else:
        current_type = d[name][0]
        # not agreement between types
        if current_type != value_type:
            # if one of the two is a string...str always wins
            if value_type == str or current_type == str:
                d[name] = (str, d[name][1])
            # between int and float...float always wins
            elif (current_type == int and value_type == float) or \
                 (current_type == float and value_type == int):
                d[name] = (float, d[name][1])
        d[name][1].add(value_type(value))



class MetadataParser:
    """ Abstract Parser of Metadata files

    """
    def parse_metadata(self, path):
        """ Parse a metadata file into a list of tuple
        [('attr_name', attr_value'), ...]

        :param path: path of the metadata file
        :return: dict
        """
        pass


class GenericMetaParser(MetadataParser):
    def __init__(self, delimiter="\t"):
        self.delimiter = delimiter

    def parse_metadata(self, path):
        fo = open(path)
        lines = fo.read().splitlines()
        return list(map(self._parse_line_metadata, lines))

    def _parse_line_metadata(self, line):
        elems = line.split(self.delimiter)
        return elems[0], elems[1]



pyType_to_scalaType = {
    "object": "string",
    "int64": "long",
    "float64": "double",
    "int32": "integer",
    "float32": "double"
}

schema_header = "<?xml version='1.0' encoding='UTF-8'?>\n"
schema_collection_template = "<gmqlSchemaCollection name= \"{}\" xmlns=\"http://genomic.elet.polimi.it/entities\">\n"
schema_dataset_type_template = "\t<gmqlSchema type=\"{}\" coordinate_system=\"{}\">\n"
schema_field_template = "\t\t<field type=\"{}\">{}</field>\n"
schema_dataset_end = "\t</gmqlSchema>\n"
schema_collection_end = "</gmqlSchemaCollection>"



from fancyimpute import KNN, SimpleFill, IterativeSVD
import pandas as pd
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import chi2, f_classif, mutual_info_classif

class Preprocessing:
    """
    Contains the functionalities for data standardization, feature selection and missing value imputation
    """
    def __init__(self):
        return

    @staticmethod
    def to_zero_mean(df):
        """
        Standardizes the data by shifting the mean value to zero

        :param df: the input dataframe
        :return: the resulting dataframe
        """
        df_norm = (df - df.mean())
        return df_norm

    @staticmethod
    def to_unit_variance(df):
        """
        Makes the variance of each gene equal to one

        :param df: the dataframe
        :return: the resulting dataframe
        """
        df_norm = df / df.std()
        return df_norm

    @staticmethod
    def prune_by_missing_percent(df, percentage=0.4):
        """
        The method to remove the attributes (genes) with more than a percentage of missing values

        :param df: the dataframe containing the attributes to be pruned
        :param percentage: the percentage threshold (0.4 by default)
        :return: the pruned dataframe
        """
        mask = (df.isnull().sum() / df.shape[0]).map(lambda x: True if x < percentage else False)
        pruned_df = df[df.columns[mask.values]]
        return pruned_df

    @staticmethod
    def impute_using_statistics(df, method='min'):
        """
        Imputes the missing values by the selected statistical property of each column

        :param df: The input dataframe that contains missing values
        :param method: The imputation method (min by default)
            "zero": fill missing entries with zeros
            "mean": fill with column means
            "median" : fill with column medians
            "min": fill with min value per column
            "random": fill with gaussian noise according to mean/std of column
        :return: the imputed dataframe
        """
        sf = SimpleFill(method)
        imputed_matrix = sf.complete(df.values)
        imputed_df = pd.DataFrame(imputed_matrix, df.index, df.columns)
        return imputed_df

    @staticmethod
    def impute_knn(df, k=3):
        """
        Nearest neighbour imputations which weights samples using the mean squared difference on features for which two rows both have observed data.

        :param df: The input dataframe that contains missing values
        :param k: The number of neighbours
        :return: the imputed dataframe
        """
        imputed_matrix = KNN(k=k).complete(df.values)
        imputed_df = pd.DataFrame(imputed_matrix, df.index, df.columns)
        return imputed_df

    @staticmethod
    def impute_svd(df, rank=10, convergence_threshold=0.00001, max_iters=200):
        """
        Imputes the missing values by using SVD decomposition
        Based on the following publication: 'Missing value estimation methods for DNA microarrays' by Troyanskaya et. al.

        :param df:The input dataframe that contains missing values
        :param rank: Rank value of the truncated SVD
        :param convergence_threshold: The threshold to stop the iterations
        :param max_iters: Max number of iterations
        :return: the imputed dataframe
        """
        imputed_matrix = IterativeSVD(rank,convergence_threshold, max_iters).complete(df.values)
        imputed_df = pd.DataFrame(imputed_matrix, df.index, df.columns)
        return imputed_df

    @staticmethod
    def feature_selection(df, labels, n_features, method='chi2'):
        """
        Reduces the number of features in the imput dataframe.
        Ex: labels = gs.meta['biospecimen_sample__sample_type_id'].apply(int).apply(lambda x: 0 if x < 10 else 1)
            chi2_fs(gs.data, labels, 50)

        :param df: The input dataframe
        :param labels: Labels for each row in the df. Type: Pandas.Series
        :param no_features: The desired number of features
        :param method: The feature selection method to be employed. It is set to 'chi2' by default
        To select the features using mutual information, the method value should be set to 'mi'
        To select the features using ANOVA, the method value should be set to 'ANOVA'
        :return: Returns the dataframe with the selected features

        """
        fs_obj = None
        if method == 'chi2':
            fs_obj = chi2
        elif method == 'ANOVA':
            fs_obj = f_classif
        elif method == 'mi':
            fs_obj = mutual_info_classif
        else:
            raise ValueError('The method is not recognized')
    
        fs = SelectKBest(fs_obj, k=n_features)
        fs.fit_transform(df, labels)
        df_reduced = df.loc[:, fs.get_support()]
        return df_reduced



"""
Biclustering algorithms.

"""

from sklearn.cluster.bicluster import SpectralBiclustering
from sklearn.cluster.bicluster import SpectralCoclustering
from sklearn.metrics import consensus_score


class Biclustering:
    """
    The class that contains the biclustering related functionalities
    """
    def __init__(self, model):
        self.model = model

    @classmethod
    def spectral_biclustering(cls, *args):
        """
        Wrapper method for the spectral_biclustering algorithm

        :param args: the arguments to be sent to the sci-kit implementation
        :return: returns the Biclustering object
        """

        model = SpectralBiclustering(*args)
        return cls(model)

    @classmethod
    def spectral_coclustering(cls, *args):
        """
        Wrapper method for the spectral_coclustering algorithm

        :param args: the arguments to be sent to the sci-kit implementation
        :return: returns the Biclustering object
        """

        model = SpectralCoclustering(*args)
        return cls(model)

    def fit(self, data):
        """
        Performs biclustering

        :param data: Data to be fit
        """
        self.model.fit(data)

    def retrieve_bicluster(self, df, row_no, column_no):
        """
        Extracts the bicluster at the given row bicluster number and the column bicluster number from the input dataframe.

        :param df: the input dataframe whose values were biclustered
        :param row_no: the number of the row bicluster
        :param column_no: the number of the column bicluster
        :return: the extracted bicluster from the dataframe
        """
        res = df[self.model.biclusters_[0][row_no]]
        bicluster = res[res.columns[self.model.biclusters_[1][column_no]]]
        return bicluster

    def bicluster_similarity(self, reference_model):
        """
        Calculates the similarity between the current model of biclusters and the reference model of biclusters

        :param reference_model: The reference model of biclusters
        :return: Returns the consensus score(Hochreiter et. al., 2010), i.e. the similarity of two sets of biclusters.
        """
        similarity_score = consensus_score(self.model.biclusters_, reference_model.biclusters_)
        return similarity_score



from pkg_resources import resource_filename
import os, shutil, time


def get_current_time():
    return time.strftime("%Y%m%d_%H%M%S")


resource_folder = resource_filename("gmql", "resources")
tmp_folder_name = os.path.join(resource_folder, "tmp")
tmp_folder_instance = os.path.join(tmp_folder_name, "Instance_" + get_current_time())
tmp_folder_spark = os.path.join(tmp_folder_instance, "spark")
tmp_folder_datasets = os.path.join(tmp_folder_instance, "datasets")

tmp_folders = [tmp_folder_name, tmp_folder_instance, tmp_folder_spark, tmp_folder_datasets]


def initialize_tmp_folders():
    for tf in tmp_folders:
        if not os.path.isdir(tf):
            os.mkdir(tf)
    result = {
        'tmp': tmp_folder_name,
        'instance': tmp_folder_instance,
        'spark': tmp_folder_spark,
        'datasets': tmp_folder_datasets
    }
    return result


def get_unique_identifier():
    return "id_" + get_current_time()


def get_new_dataset_tmp_folder():
    name = get_unique_identifier()
    return os.path.join(tmp_folder_datasets, name)


def delete_tmp_dataset(path):
    if os.path.isdir(path):
        shutil.rmtree(path)


def flush_tmp_folder(folder):
    if os.path.isdir(folder):
        shutil.rmtree(folder)


def flush_everything():
    flush_tmp_folder(tmp_folder_instance)
    # check if the top folder is empty or not
    if not os.listdir(tmp_folder_name):
        flush_tmp_folder(tmp_folder_name)




from .. import GDataframe
from tqdm import tqdm
import logging
from . import *
import os
import shutil
from ..parsers import TAB, COORDS_DEFAULT

logger = logging.getLogger("PyGML logger")


def to_dataset_files(gframe, path_local=None, path_remote=None):
    if not isinstance(gframe, GDataframe.GDataframe):
        raise TypeError("Expected GDataframe, got {}".format(type(gframe)))

    if not check_gmql_coherent(gframe):
        raise ValueError("The GDataframe is not coherent with the GMQL standard")

    if path_local is not None:
        to_local(gframe, path_local)

    if path_remote is not None:
        to_remote(gframe, path_remote)


def to_local(gframe, path_local):
    regs = gframe.regs
    meta = gframe.meta

    # creating the directory
    if os.path.isdir(path_local):
        shutil.rmtree(path_local)
    os.makedirs(path_local)

    logger.info("Saving data at {}".format(path_local))
    all_file_names = regs.index.unique()
    from ... import __disable_progress
    for file in tqdm(all_file_names, disable=__disable_progress):
        final_file_name = os.path.join(path_local, str(file))
        part_regs = regs.loc[[file]]
        part_regs.to_csv(path_or_buf=final_file_name, sep="\t", header=False, index=False)

        parts_meta = meta.loc[[file]]
        meta_to_file(parts_meta, final_file_name + ".meta")

    # save the schema
    dataset_name = os.path.basename(os.path.normpath(path_local))
    save_schema(regs, dataset_name, TAB, os.path.join(path_local, "schema.schema"), COORDS_DEFAULT)


def meta_to_file(meta_df, filename):
    f = open(filename, "w")
    # the dataframe is only one line
    d = meta_df.iloc[0].to_dict()

    for k in d.keys():
        line_template = "{}\t{}\n"
        data = d[k]
        for el in data:
            line = line_template.format(k, el)
            f.write(line)
    f.close()


def save_schema(regs_df, dataset_name, dataset_type, schema_file_name, coordinate_system):
    column_names = regs_df.columns.tolist()
    types = regs_df.dtypes
    result = schema_header
    result += schema_collection_template.format(dataset_name)
    result += schema_dataset_type_template.format(dataset_type, coordinate_system)
    for c in column_names:
        t = pyType_to_scalaType[types[c].name].upper()
        result += schema_field_template.format(t, c)
    result += schema_dataset_end
    result += schema_collection_end

    with open(schema_file_name, "w") as f:
        f.write(result)


def to_remote(gframe, path_remote):
    pass


def check_gmql_coherent(gframe):
    # index coherency
    regs_idxs = set(gframe.regs.index.unique())
    meta_idxs = set(gframe.meta.index.unique())
    if regs_idxs != meta_idxs:
        return False
    return True





import pandas as pd
import os
import xml.etree.ElementTree
from tqdm import tqdm


class Parser:
    """

    General purpose parser for the (tab separated) output of the MAP operations of GMQL.
    This parser allows the parsing of a subset of region and meta data of interest to improve performance

    """

    def __init__(self, path):
        self.path = path
        self.schema = self._get_schema_file("schema", path)
        return

    @staticmethod
    def get_sample_id(path):
        """
        Returns the id of the sample (the filename) file given a path
        :param path: Path of the sample file
        :return: The resulting sample id
        """
        sp = path.split('/')
        file_name = sp[-1]
        return file_name.split('.')[0]

    @staticmethod
    def _get_files(extension, path):
        """
        Returns a sorted list of all of the files having the same extension under the same directory
        :param extension: the extension of the data files such as 'gdm'
        :param path: path to the folder containing the files
        :return: sorted list of files
        """
        # retrieves the files sharing the same extension
        files = []
        for file in os.listdir(path):
            if file.endswith(extension):
                files.append(os.path.join(path, file))
        return sorted(files)

    @staticmethod
    def _get_schema_file(extension, path):
        """
        Returns the schema file
        :param extension: extension of the schema file usually .schema
        :param path: path to the folder containing the schema file
        :return: the path to the schema file
        """
        for file in os.listdir(path):
            if file.endswith(extension):
                return os.path.join(path, file)

    @staticmethod
    def parse_schema(schema_file):
        """
        parses the schema file and returns the columns that are later going to represent the columns of the genometric space dataframe
        :param schema_file: the path to the schema file
        :return: the columns of the schema file
        """
        e = xml.etree.ElementTree.parse(schema_file)
        root = e.getroot()
        cols = []
        for elem in root.findall(".//{http://genomic.elet.polimi.it/entities}field"):  # XPATH
            cols.append(elem.text)
        return cols

    @staticmethod
    def _get_sample_name(path):
        """
        Returns the sample name given the path to the sample file
        :param path: path to the sample file
        :return: name of the sample
        """
        sp = path.split('/')
        file_name = sp[-1]
        return file_name.split('.')[0]

    def parse_single_meta(self, fname, selected_meta_data):
        """
        Parses a single meta data file
        :param fname: name of the file
        :param selected_meta_data: If not none then only the specified columns of metadata are parsed
        :return: the resulting pandas series
        """
        # reads a meta data file into a dataframe
        columns = []
        data = []
        with open(fname) as f:
            for line in f:
                splitted = line.split('\t')
                columns.append(splitted[0])
                data.append(splitted[1].split('\n')[0])  # to remove the \n values
        df = pd.DataFrame(data=data, index=columns)
        df = df.T
        sample = self._get_sample_name(fname)
        if selected_meta_data:  # if null then keep all the columns
            try:
                df = df[selected_meta_data]
            except:
                pass
        df['sample'] = sample
        return df

    def parse_meta(self, selected_meta_data):
        """
        Parses all of the metadata files
        :param selected_meta_data: if specified then only the columns that are contained here are going to be parsed
        :return:
        """
        # reads all meta data files
        files = self._get_files("meta", self.path)
        df = pd.DataFrame()
        print("Parsing the metadata files...")
        for f in tqdm(files):
            data = self.parse_single_meta(f, selected_meta_data)
            if data is not None:
                df = pd.concat([df, data], axis=0)
        df.index = df['sample']
        #
        # df = df.drop('sample', 1)  # 1 for the columns
        return df

    def parse_single_data(self, path, cols, selected_region_data, selected_values, full_load):
        """
        Parses a single region data file
        :param path: path to the file
        :param cols: the column names coming from the schema file
        :param selected_region_data: the selected of the region data to be parsed
        In most cases the analyst only needs a small subset of the region data
        :param selected_values: the selected values to be put in the matrix cells
        :param full_load: Specifies the method of parsing the data. If False then parser omits the parsing of zero(0)
            values in order to speed up and save memory. However, while creating the matrix, those zero values are going to be put into the matrix.
            (unless a row contains "all zero columns". This parsing is strongly recommended for sparse datasets.
            If the full_load parameter is True then all the zero(0) data are going to be read.
        :return: the dataframe containing the region data
        """
        # reads a sample file
        df = pd.read_table(path, engine='c', sep="\t", lineterminator="\n", header=None)
        df.columns = cols  # column names from schema
        df = df[selected_region_data]

        if not full_load:
            if type(selected_values) is list:
                df_2 = pd.DataFrame(dtype=float)
                for value in selected_values:
                    df_3 = df.loc[df[value] != 0]
                    df_2 = pd.concat([df_2, df_3], axis=0)
                df = df_2
            else:
                df = df.loc[df[selected_values] != 0]

        sample = self._get_sample_name(path)
        df['sample'] = sample
        return df

    def parse_data(self, selected_region_data, selected_values, full_load=False, extension="gdm"):
        """
        Parses all of the region data
        :param selected_region_data: the columns of region data that are needed
        :param selected_values: the selected values to be put in the matrix cells
        :param full_load: Specifies the method of parsing the data. If False then parser omits the parsing of zero(0)
            values in order to speed up and save memory. However, while creating the matrix, those zero values are going to be put into the matrix.
            (unless a row contains "all zero columns". This parsing is strongly recommended for sparse datasets.
            If the full_load parameter is True then all the zero(0) data are going to be read.
        :param extension: the extension of the region data files that are going to be parsed.
        :return: the resulting region dataframe
        """
        regions = list(selected_region_data)
        if type(selected_values) is list:
            regions.extend(selected_values)
        else:
            regions.append(selected_values)

        files = self._get_files(extension, self.path)
        df = pd.DataFrame(dtype=float)

        cols = self.parse_schema(self.schema)
        print("Parsing the data files...")
        list_of_data = []
        for f in tqdm(files):
            data = self.parse_single_data(f, cols, regions, selected_values, full_load)
            list_of_data.append(data)
        print("pre-concat")
        df = pd.concat(list_of_data)
        return df



"""
Clustering algorithms

"""

from sklearn.cluster import *
from sklearn.metrics.cluster import *
from pyclustering.cluster.xmeans import xmeans, splitting_type
from pyclustering.cluster.clarans import clarans
from pyclustering.cluster.rock import rock
from pyclustering.cluster.optics import optics
from scipy.spatial.distance import cdist, pdist
from matplotlib import pyplot as plt

import pandas as pd
import numpy as np

class Clustering:
    """
    The class that contains the algorithms and related methods
    """
    def __init__(self, model):
        self.model = model

    @classmethod
    def xmeans(cls, initial_centers=None, kmax=20, tolerance=0.025, criterion=splitting_type.BAYESIAN_INFORMATION_CRITERION, ccore=False):
        """
        Constructor of the x-means clustering.rst algorithm

        :param initial_centers: Initial coordinates of centers of clusters that are represented by list: [center1, center2, ...]
        Note: The dimensions of the initial centers should be same as of the dataset.
        :param kmax: Maximum number of clusters that can be allocated.
        :param tolerance: Stop condition for each iteration: if maximum value of change of centers of clusters is less than tolerance than algorithm will stop processing
        :param criterion: Type of splitting creation.
        :param ccore: Defines should be CCORE (C++ pyclustering library) used instead of Python code or not.
        :return: returns the clustering.rst object
        """
        model = xmeans(None, initial_centers, kmax, tolerance, criterion, ccore)
        return cls(model)

    @classmethod
    def clarans(cls, number_clusters, num_local, max_neighbour):
        """
        Constructor of the CLARANS clustering.rst algorithm

        :param number_clusters: the number of clusters to be allocated
        :param num_local: the number of local minima obtained (amount of iterations for solving the problem).
        :param max_neighbour: the number of local minima obtained (amount of iterations for solving the problem).
        :return: the resulting clustering.rst object
        """
        model = clarans(None, number_clusters, num_local, max_neighbour)
        return cls(model)

    @classmethod
    def rock(cls, data, eps, number_clusters, threshold=0.5, ccore=False):
        """
        Constructor of the ROCK cluster analysis algorithm

        :param eps: Connectivity radius (similarity threshold), points are neighbors if distance between them is less than connectivity radius
        :param number_clusters: Defines number of clusters that should be allocated from the input data set
        :param threshold: Value that defines degree of normalization that influences on choice of clusters for merging during processing
        :param ccore: Defines should be CCORE (C++ pyclustering library) used instead of Python code or not.
        :return: The resulting clustering.rst object
        """
        data = cls.input_preprocess(data)
        model = rock(data, eps, number_clusters, threshold, ccore)
        return cls(model)

    @staticmethod
    def input_preprocess(data):
        if isinstance(data, pd.DataFrame):
            data = data.values.tolist()
        elif isinstance(data, np.ndarray):  # in case data is already in the matrix form
            data = data.tolist()
        return data

    @classmethod
    def optics(cls, data, eps, minpts, ccore=False):
        """
        Constructor of OPTICS clustering.rst algorithm

        :param data: Input data that is presented as a list of points (objects), where each point is represented by list or tuple
        :param eps: Connectivity radius between points, points may be connected if distance between them less than the radius
        :param minpts: Minimum number of shared neighbors that is required for establishing links between points
        :param amount_clusters: Optional parameter where amount of clusters that should be allocated is specified.
                    In case of usage 'amount_clusters' connectivity radius can be greater than real, in other words, there is place for mistake
                    in connectivity radius usage.
        :param ccore: if True than DLL CCORE (C++ solution) will be used for solving the problem
        :return: the resulting clustering.rst object
        """
        data = cls.input_preprocess(data)
        model = optics(data, eps, minpts)
        return cls(model)


    @classmethod
    def kmeans(cls, *args):
        """
            Constructor of the k-means clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = KMeans(*args)
        return cls(model)

    @classmethod
    def affinity_propagation(cls, *args):
        """
            Constructor of the affinity propagation clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = AffinityPropagation(*args)
        return cls(model)

    @classmethod
    def hierarchical(cls, *args):
        """
            Constructor of the agglomerative clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = AgglomerativeClustering(*args)
        return cls(model)

    @classmethod
    def birch(cls, *args):
        """
            Constructor of the birch clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = Birch(*args)
        return cls(model)

    @classmethod
    def dbscan(cls, *args):
        """
            Constructor of the DBSCAN clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = DBSCAN(*args)
        return cls(model)

    @classmethod
    def feature_agglomeration(cls, *args):
        """
            Constructor of the feature agglomeration clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = FeatureAgglomeration(*args)
        return cls(model)

    @classmethod
    def mini_batch_kmeans(cls, *args):
        """
            Constructor of the mini batch k-means clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = MiniBatchKMeans(*args)
        return cls(model)

    @classmethod
    def mean_shift(cls, *args):
        """
            Constructor of the mean shift clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = MeanShift(*args)
        return cls(model)

    @classmethod
    def spectral_clustering(cls, *args):
        """
            Constructor of the spectral clustering.rst algorithm

            :param args: the arguments to be sent to the sci-kit implementation
            :return: returns the Clustering object
        """

        model = SpectralClustering(*args)
        return cls(model)

    @staticmethod
    def is_pyclustering_instance(model):
        """
        Checks if the clustering.rst algorithm belongs to pyclustering

        :param model: the clustering.rst algorithm model
        :return: the truth value (Boolean)
        """
        return any(isinstance(model, i) for i in [xmeans, clarans, rock, optics])

    def fit(self, data=None):
        """
        Performs clustering.rst

        :param data: Data to be fit
        :return: the clustering.rst object
        """
        if self.is_pyclustering_instance(self.model):
            if isinstance(self.model, xmeans):
                data = self.input_preprocess(data)
                self.model._xmeans__pointer_data = data
            elif isinstance(self.model, clarans):
                data = self.input_preprocess(data)
                self.model._clarans__pointer_data = data

            self.model.process()
        else:
            self.model.fit(data)
            return self

    @property
    def _labels_from_pyclusters(self):
        """
        Computes and returns the list of labels indicating the data points and the corresponding cluster ids.

        :return: The list of labels
        """
        clusters = self.model.get_clusters()
        labels = []
        for i in range(0, len(clusters)):
            for j in clusters[i]:
                labels.insert(int(j), i)
        return labels

    def retrieve_cluster(self, df, cluster_no):
        """
        Extracts the cluster at the given index from the input dataframe

        :param df: the dataframe that contains the clusters
        :param cluster_no: the cluster number
        :return: returns the extracted cluster
        """
        if self.is_pyclustering_instance(self.model):
            clusters = self.model.get_clusters()
            mask = []
            for i in range(0, df.shape[0]):
                mask.append(i in clusters[cluster_no])
        else:
            mask = self.model.labels_ == cluster_no  # a boolean mask
        return df[mask]

    @staticmethod
    def get_labels(obj):
        """
        Retrieve the labels of a clustering.rst object

        :param obj: the clustering.rst object
        :return: the resulting labels
        """
        if Clustering.is_pyclustering_instance(obj.model):
            return obj._labels_from_pyclusters
        else:
            return obj.model.labels_

    @staticmethod
    def silhouette_n_clusters(data, k_min, k_max, distance='euclidean'):
        """
        Computes and plot the silhouette score vs number of clusters graph to help selecting the number of clusters visually

        :param data: The data object
        :param k_min: lowerbound of the cluster range
        :param k_max: upperbound of the cluster range
        :param distance: the distance metric, 'euclidean' by default
        :return:
        """
        k_range = range(k_min, k_max)

        k_means_var = [Clustering.kmeans(k).fit(data) for k in k_range]

        silhouette_scores = [obj.silhouette_score(data=data, metric=distance) for obj in k_means_var]

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(k_range, silhouette_scores, 'b*-')
        ax.set_ylim((-1, 1))
        plt.grid(True)
        plt.xlabel('n_clusters')
        plt.ylabel('The silhouette score')
        plt.title('Silhouette score vs. k')
        plt.show()


    @staticmethod
    def elbow_method(data, k_min, k_max, distance='euclidean'):
        """
        Calculates and plots the plot of variance explained - number of clusters
        Implementation reference: https://github.com/sarguido/k-means-clustering.rst

        :param data: The dataset
        :param k_min: lowerbound of the cluster range
        :param k_max: upperbound of the cluster range
        :param distance: the distance metric, 'euclidean' by default
        :return:
        """
        # Determine your k range
        k_range = range(k_min, k_max)

        # Fit the kmeans model for each n_clusters = k
        k_means_var = [Clustering.kmeans(k).fit(data) for k in k_range]

        # Pull out the cluster centers for each model
        centroids = [X.model.cluster_centers_ for X in k_means_var]

        # Calculate the Euclidean distance from
        # each point to each cluster center
        k_euclid = [cdist(data, cent, distance) for cent in centroids]
        dist = [np.min(ke, axis=1) for ke in k_euclid]

        # Total within-cluster sum of squares
        wcss = [sum(d ** 2) for d in dist]

        # The total sum of squares
        tss = sum(pdist(data) ** 2) / data.shape[0]

        # The between-cluster sum of squares
        bss = tss - wcss

        # elbow curve
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(k_range, bss / tss * 100, 'b*-')
        ax.set_ylim((0, 100))
        plt.grid(True)
        plt.xlabel('n_clusters')
        plt.ylabel('Percentage of variance explained')
        plt.title('Variance Explained vs. k')
        plt.show()

    def adjusted_mutual_info(self, reference_clusters):
        """
        Calculates the adjusted mutual information score w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: returns the value of the adjusted mutual information score
        """
        return adjusted_mutual_info_score(self.get_labels(self), self.get_labels(reference_clusters))

    def adjusted_rand_score(self, reference_clusters):
        """
        Calculates the adjusted rand score w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: returns the value of the adjusted rand score
        """
        return adjusted_rand_score(self.get_labels(self), self.get_labels(reference_clusters))

    def calinski_harabasz(self, data):
        """
        Calculates the Calinski-Harabarsz score for a set of clusters (implicit evaluation).

        :param data: The original dataset that the clusters are generated from
        :return: The resulting Calinski-Harabarsz score
        """
        return calinski_harabaz_score(data, self.get_labels(self))

    def completeness_score(self, reference_clusters):
        """
        Calculates the completeness score w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: the resulting completeness score
        """
        return completeness_score(self.get_labels(self), self.get_labels(reference_clusters))

    def fowlkes_mallows(self, reference_clusters):
        """
        Calculates the Fowlkes-Mallows index (FMI) w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: The resulting Fowlkes-Mallows score.
        """
        return fowlkes_mallows_score(self.get_labels(self), self.get_labels(reference_clusters))

    def homogeneity_score(self, reference_clusters):
        """
        Calculates the homogeneity score w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: The resulting homogeneity score.
        """
        return homogeneity_score(self.get_labels(self), self.get_labels(reference_clusters))

    def mutual_info_score(self, reference_clusters):
        """
        Calculates the MI (mutual information) w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: The resulting MI score.
        """
        return mutual_info_score(self.get_labels(self), self.get_labels(reference_clusters))

    def normalized_mutual_info_score(self, reference_clusters):
        """
        Calculates the normalized mutual information w.r.t. the reference clusters (explicit evaluation)

        :param reference_clusters: Clusters that are to be used as reference
        :return: The resulting normalized mutual information score.
        """
        return normalized_mutual_info_score(self.get_labels(self), self.get_labels(reference_clusters))

    def silhouette_score(self, data,  metric='euclidean', sample_size=None, random_state=None, **kwds):
        """
        Computes the mean Silhouette Coefficient of all samples (implicit evaluation)

        :param data: The data that the clusters are generated from
        :param metric: the pairwise distance metric
        :param sample_size: the size of the sample to use computing the Silhouette Coefficient
        :param random_state: If an integer is given then it fixes its seed otherwise random.
        :param kwds: any further parameters that are passed to the distance function
        :return: the mean Silhouette Coefficient of all samples
        """
        return silhouette_score(data, self.get_labels(self), metric, sample_size, random_state, **kwds)



from . import default_address, headers
from .. import get_python_manager, is_progress_enabled
import requests, time, logging, json
from requests_toolbelt.multipart.encoder import MultipartEncoderMonitor, MultipartEncoder
import pandas as pd
from ..dataset.GDataframe import GDataframe
from ..dataset.parsers.RegionParser import RegionParser
from ..dataset.parsers import GTF
from ..dataset.DataStructures import chr_aliases, start_aliases, stop_aliases, strand_aliases
from ..dataset.loaders import Loader
from ..FileManagment import TempFileManager
import os, zipfile
from tqdm import tqdm
from ..dataset.storers.parserToXML import parserToXML
import warnings
import threading
import numpy as np


good_status = ['PENDING', 'RUNNING', 'DS_CREATION_RUNNING']
CHUNK_SIZE = 5 * 1024 * 1024  # 5 MB


class RemoteManager:
    """ Manager of the user connection with the remote GMQL service
    """
    def __init__(self, address=None, auth_token=None):
        """ Instantiate a new RemoteManager. If no address is provided, the default address will
        be used, which represents the current server hosted by the Dept. of Electronics, Information
        and Bioengineering of Politecnico di Milano.

        :param address: (optional) the address of the remote GMQL service
        """

        # checking of the address
        if address is None:
            self.address = default_address
        elif isinstance(address, str):
            address = address.strip()
            if address.endswith("/"):
                self.address = address[:-1]
            else:
                self.address = address
        else:
            raise TypeError("The remote URL must be a string."
                            " {} was provided".format(type(address)))

        # checking the existence of the remote service
        req = requests.get(self.address + "/")
        if req.status_code != 200:
            raise ConnectionError("The server at {} is not responding".format(self.address))

        self.logger = logging.getLogger("PyGML logger")

        # checking of the auth_token
        if auth_token is not None:
            header = self.__get_header(auth_token)
            url = self.address + "/datasets"
            response = requests.get(url, headers=header)
            if response.status_code == 200:
                # the auth_token is valid
                # print("VALID AUTH TOKEN")
                self.auth_token = auth_token
            elif response.status_code == 401 and \
                response.json().get("error") == 'UnAuthenticatedRequest':
                # print("NOT-VALID AUTH TOKEN")
                self.auth_token = None
            else:
                raise ValueError("Unknown state. {} ".format(response.status_code))
        else:
            self.auth_token = None
        self.json_encoder = json.JSONEncoder()

    """
        Security controls
    """

    def register(self, first_name, last_name, user_name, email, password):
        url = self.address + "/register"
        body = {
            "firstName" : first_name,
            "lastName": last_name,
            "username": user_name,
            "email": email,
            "password": password
        }
        response = requests.post(url, data=json.dumps(body))
        if response.status_code != 200:
            raise ValueError("Code {}. {}".format(response.status_code, response.json().get("error")))

    def login(self, username=None, password=None):
        """ Before doing any remote operation, the user has to login to the GMQL serivice.
        This can be done in the two following ways:

            * Guest mode: the user has no credentials and uses the system only as a temporary guest
            * Authenticated mode: the users has credentials and a stable remote account

        If neither username and password are specified, the user enters the system as a guest.
        If both are specified and they correspond to an existent user, the user enters as an
        authenticated user

        :param username: (optional)
        :param password: (optional)
        :return: None
        """
        if (username is None) and (password is None):
            auth_token = self.__login_guest()
        elif (username is not None) and (password is not None):
            auth_token, fullName = self.__login_credentials(username, password)
            self.logger.info("You are logged as {}".format(fullName))
        else:
            raise ValueError("you have to specify both username and password or nothing")

        if auth_token is not None:
            self.auth_token = auth_token
        else:
            raise ConnectionError("Impossible to retrieve the authentication token")

    def auto_login(self, how="guest"):
        if self.auth_token is None:
            if how != 'guest':
                warnings.warn("The authentication token for your account is expired. "
                              "You need to redo the login using the pygmql tool. "
                              "For now you will be logged as guest user")
            self.auth_token = self.__login_guest()
            return "guest"
        else:
            return how

    def __login_guest(self):
        url = self.address + "/guest"
        response = requests.get(url=url, headers=headers)
        response = response.json()
        return response.get("authToken")

    def __login_credentials(self, username, password):
        url = self.address + "/login"
        body = {
            "username": username,
            "password": password
        }
        # body = self.json_encoder.encode(body)
        response = requests.post(url, data=json.dumps(body), headers=headers)
        response = response.json()

        errorString = response.get("errorString")
        if errorString is not None:
            raise ValueError(errorString)
        else:
            auth_token = response.get("authToken")
            fullName = response.get("fullName")
        return auth_token, fullName

    @staticmethod
    def __get_header(auth_token):
        header = headers.copy()
        header['X-AUTH-TOKEN'] = auth_token
        return header

    def __check_authentication(self):
        if self.auth_token is not None:
            return self.__get_header(self.auth_token)
        else:
            raise EnvironmentError("you first need to login before doing operations")

    def logout(self):
        """ Logout from the remote account

        :return: None
        """
        url = self.address + "/logout"
        header = self.__check_authentication()
        response = requests.get(url, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}. {}".format(response.status_code, response.json().get("error")))

    """
        Repository
    """

    @staticmethod
    def process_info_list(res, info_column):
        def extract_infos(row):
            infoList = row['infoList']
            result = {}
            for d in infoList:
                result[d['key']] = d['value']
            return result
        res = pd.concat([res, pd.DataFrame.from_dict(res[info_column].map(extract_infos).tolist())], axis=1)\
            .drop("info", axis=1)
        return res

    def get_dataset_list(self):
        """ Returns the list of available datasets for the current user.

        :return: a pandas Dataframe
        """
        url = self.address + "/datasets"
        header = self.__check_authentication()
        response = requests.get(url, headers=header)
        response = response.json()
        datasets = response.get("datasets")
        res = pd.DataFrame.from_dict(datasets)
        return self.process_info_list(res, "info")

    def get_dataset_samples(self, dataset_name, owner=None):
        """ Get the list of samples of a specific remote dataset.

        :param dataset_name: the dataset name
        :param owner: (optional) who owns the dataset. If it is not specified, the current user
               is used. For public dataset use 'public'.
        :return: a pandas Dataframe
        """
        if isinstance(owner, str):
            owner = owner.lower()
            dataset_name = owner + "." + dataset_name

        header = self.__check_authentication()

        url = self.address + "/datasets/" + dataset_name
        response = requests.get(url, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}: {}".format(response.status_code, response.json().get("error")))
        response = response.json()
        samples = response.get("samples")
        if len(samples) == 0:
            return None
        res = pd.DataFrame.from_dict(samples)
        return self.process_info_list(res, "info")

    def get_dataset_schema(self, dataset_name, owner=None):
        """ Given a dataset name, it returns a BedParser coherent with the schema of it

        :param dataset_name: a dataset name on the repository
        :param owner: (optional) who owns the dataset. If it is not specified, the current user
               is used. For public dataset use 'public'.
        :return: a BedParser
        """

        if isinstance(owner, str):
            owner = owner.lower()
            dataset_name = owner + "." + dataset_name

        url = self.address + "/datasets/" + dataset_name+"/schema"
        header = self.__check_authentication()
        response = requests.get(url, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}: {}".format(response.status_code, response.json().get("error")))

        response = response.json()
        name = response.get("name")
        schemaType = response.get("type")
        coordinates_system = response.get("coordinate_system")
        fields = response.get("fields")

        i = 0
        chrPos, startPos, stopPos, strandPos = None, None, None, None
        otherPos = []
        if schemaType == GTF:
            chrPos = 0  # seqname
            startPos = 3  # start
            stopPos = 4  # end
            strandPos = 6  # strand
            otherPos = [(1, 'source', 'string'), (2, 'feature', 'string'),
                        (5, 'score', 'float'), (7, 'frame', 'string')]

            for field in fields:
                fieldName = field.get("name").lower()
                fieldType = field.get("type").lower()
                if fieldName not in {'seqname', 'start', 'end', 'strand',
                                     'source', 'feature', 'score', 'frame'}:
                    otherPos.append((i, fieldName, fieldType))
                i += 1

        else:
            for field in fields:
                fieldName = field.get("name").lower()
                fieldType = field.get("type").lower()

                if fieldName in chr_aliases and chrPos is None:
                    chrPos = i
                elif fieldName in start_aliases and startPos is None:
                    startPos = i
                elif fieldName in stop_aliases and stopPos is None:
                    stopPos = i
                elif fieldName in strand_aliases and strandPos is None:
                    strandPos = i
                else:  # other positions
                    otherPos.append((i, fieldName, fieldType))
                i += 1
        if len(otherPos) == 0:
            otherPos = None

        return RegionParser(chrPos=chrPos,
                            startPos=startPos,
                            stopPos=stopPos,
                            strandPos=strandPos,
                            otherPos=otherPos,
                            schema_format=schemaType,
                            coordinate_system=coordinates_system,
                            delimiter="\t", parser_name=name)

    def upload_dataset(self, dataset, dataset_name, schema_path=None):
        """ Upload to the repository an entire dataset from a local path

        :param dataset: the local path of the dataset
        :param dataset_name: the name you want to assign to the dataset remotely
        :return: None
        """

        url = self.address + "/datasets/" + dataset_name + "/uploadSample"
        header = self.__check_authentication()

        fields = dict()
        remove = False
        if isinstance(dataset, GDataframe):
            tmp_path = TempFileManager.get_new_dataset_tmp_folder()
            dataset.to_dataset_files(local_path=tmp_path)
            dataset = tmp_path
            remove = True

        # a path is provided
        if not isinstance(dataset, str):
            raise TypeError("Dataset can be a path or a GDataframe. {} was passed".format(type(dataset)))

        file_paths, schema_path_found = Loader.get_file_paths(dataset)
        if schema_path is None:
            schema_path = schema_path_found
        fields['schema'] = (os.path.basename(schema_path), open(schema_path, "rb"), 'application/octet-stream')
        for i, file in enumerate(file_paths):
            fields["file"+str(i + 1)] = (os.path.basename(file), open(file, "rb"), 'application/octet-stream')

        encoder = MultipartEncoder(fields)
        callback = create_callback(encoder, len(fields))

        m_encoder = MultipartEncoderMonitor(encoder, callback)
        header['Content-Type'] = m_encoder.content_type

        self.logger.info("Uploading dataset at {} with name {}".format(dataset, dataset_name))

        response = requests.post(url, data=m_encoder,
                                 headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}: {}".format(response.status_code, response.content))

        if remove:
            TempFileManager.delete_tmp_dataset(dataset)

    def delete_dataset(self, dataset_name):
        """ Deletes the dataset having the specified name

        :param dataset_name: the name that the dataset has on the repository
        :return: None
        """
        url = self.address + "/datasets/" + dataset_name
        header = self.__check_authentication()
        response = requests.delete(url, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}: {}".format(response.status_code, response.json().get("error")))
        self.logger.info("Dataset {} was deleted from the repository".format(dataset_name))

    """
        Download repository
    """

    def download_dataset(self, dataset_name, local_path, how="stream"):
        """ It downloads from the repository the specified dataset and puts it
        in the specified local folder

        :param dataset_name: the name the dataset has in the repository
        :param local_path: where you want to save the dataset
        :param how: 'zip' downloads the whole dataset as a zip file and decompress it; 'stream'
               downloads the dataset sample by sample
        :return: None
        """
        if not os.path.isdir(local_path):
            os.makedirs(local_path)

        if how == 'zip':
            return self.download_as_zip(dataset_name, local_path)
        elif how == 'stream':
            return self.download_as_stream(dataset_name, local_path)
        else:
            raise ValueError("how must be {'zip', 'stream'}")

    def download_as_zip(self, dataset_name, local_path):
        header = self.__check_authentication()
        url = self.address + "/datasets/" + dataset_name + "/zip"
        self.logger.info("Downloading dataset {} to {}".format(dataset_name, local_path))
        response = requests.get(url, stream=True, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}: {}".format(response.status_code, response.json().get("error")))
        tmp_zip = os.path.join(local_path, "tmp.zip")
        f = open(tmp_zip, "wb")
        total_size = int(response.headers.get("content-length", 0))
        # TODO: find a better way to display the download progression
        for chunk in tqdm(response.iter_content(chunk_size=CHUNK_SIZE), total=total_size/CHUNK_SIZE,
                          disable=not is_progress_enabled(), unit="B", unit_scale=True):
            if chunk:
                f.write(chunk)
        f.close()
        with zipfile.ZipFile(tmp_zip, "r") as zip_ref:
            zip_ref.extractall(local_path)
        os.remove(tmp_zip)

    def download_as_stream(self, dataset_name, local_path):

        N_THREADS = 10

        def thread_download(sample_names):
            for sn in sample_names:
                self.download_sample(dataset_name=dataset_name,
                                     sample_name=sn,
                                     local_path=local_path,
                                     how="all",
                                     header=False)
                pbar.update()

        samples = self.get_dataset_samples(dataset_name)
        if samples is not None:
            threads = []
            ids = samples.id.unique()
            pbar = tqdm(total=len(ids), disable=not is_progress_enabled())
            splits = np.array_split(ids, N_THREADS)
            for ssn in splits:
                names = samples[samples.id.isin(ssn)].name.values
                t = threading.Thread(target=thread_download, args=(names, ))
                t.start()
                threads.append(t)
            for t in threads:
                t.join()
            pbar.close()
        schema = self.get_dataset_schema(dataset_name=dataset_name)
        parserToXML(parser=schema, datasetName=dataset_name, path=os.path.join(local_path, dataset_name + ".schema"))

    def download_sample(self, dataset_name, sample_name, local_path, how="all", header=False):
        header_get = self.__check_authentication()
        url = self.address + "/datasets/{}/{}/{}?header={}"
        region = False
        meta = False

        sample_path = os.path.join(local_path, sample_name)
        region_path = sample_path
        meta_path = sample_path + ".meta"

        if how == 'regs':
            region = True
        elif how == 'meta':
            meta = True
        elif how == 'all':
            region = True
            meta = True
        else:
            raise ValueError("how must be {'regs', 'meta', 'all'}")

        header = "true" if header else "false"

        if region:
            url_region = url.format(dataset_name, sample_name, "region", header)
            response = requests.get(url_region, stream=True, headers=header_get)
            with open(region_path, "wb") as f:
                for data in response.iter_content(chunk_size=CHUNK_SIZE):
                    f.write(data)

        if meta:
            url_meta = url.format(dataset_name, sample_name, "metadata", header)
            response = requests.get(url_meta, stream=True, headers=header_get)
            with open(meta_path, "wb") as f:
                for data in response.iter_content(chunk_size=CHUNK_SIZE):
                    f.write(data)

    """
        Query
    """

    def query(self, query, output_path=None, file_name="query", output="tab"):
        """ Execute a GMQL textual query on the remote server.

        :param query: the string containing the query
        :param output_path (optional): where to store the results locally. If specified
               the results are downloaded locally
        :param file_name (optional): the name of the query
        :param output (optional): how to save the results. It can be "tab" or "gtf"
        :return: a pandas dataframe with the dictionary ids of the results
        """
        header = self.__check_authentication()
        header['Content-Type'] = "text/plain"
        output = output.lower()
        if output not in ['tab', 'gtf']:
            raise ValueError("output must be 'tab' or 'gtf'")
        url = self.address + "/queries/run/" + file_name + '/' + output
        response = requests.post(url, data=query, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}. {}".format(response.status_code, response.json().get("error")))
        response = response.json()
        jobid = response.get("id")
        self.logger.info("JobId: {}. Waiting for the result".format(jobid))

        status_resp = self._wait_for_result(jobid)

        datasets = status_resp.get("datasets")
        result = []
        for dataset in datasets:
            name = dataset.get("name")
            result.append({'dataset': name})
            if output_path is not None:
                path = os.path.join(output_path, name)
                self.download_dataset(dataset_name=name, local_path=path)
        return pd.DataFrame.from_dict(result)

    def _wait_for_result(self, jobid):
        count = 1
        while True:
            status_resp = self.trace_job(jobid)
            status = status_resp["status"]
            if status == 'SUCCESS':
                break
            elif status in good_status:
                print(" "*50, end='\r')
                dots = "." * (count % 4)
                print(status + dots, end="\r")
            else:
                message = status_resp['message']
                raise ValueError("Status: {}. Error during query execution: {}"
                                 .format(status, message))
            count += 1
            time.sleep(1)
        return status_resp

    # def execute_remote(self, dataset, output="tab"):
    #     if not isinstance(dataset, GMQLDataset):
    #         raise TypeError("A GMQLDataset is required. {} was found".format(type(dataset)))
    #
    #     dag = dataset._get_serialized_dag()
    #     self._execute_dag(dag, output)
    #     # TODO: complete...

    def execute_remote_all(self, output="tab", output_path=None):
        pmg = get_python_manager()
        serialized_dag = pmg.get_serialized_materialization_list()
        pmg.getServer().clearMaterializationList()
        return self._execute_dag(serialized_dag, output, output_path)

    def _execute_dag(self, serialized_dag, output="tab", output_path=None):
        header = self.__check_authentication()
        header['Content-Type'] = "text/plain"
        output = output.lower()
        if output not in ['tab', 'gtf']:
            raise ValueError("output must be 'tab' or 'gtf'")

        url = self.address + "/queries/dag/" + output
        body = serialized_dag
        response = requests.post(url=url, data=body, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}. {}".format(response.status_code, response.json().get("error")))
        response = response.json()
        jobid = response.get("id")
        self.logger.info("JobId: {}. Waiting for the result".format(jobid))
        status_resp = self._wait_for_result(jobid)

        datasets = status_resp.get("datasets")
        result = []

        if isinstance(output_path, bool):
            if output_path:
                output_path = TempFileManager.get_new_dataset_tmp_folder()
            else:
                output_path = None

        for dataset in datasets:
            name = dataset.get("name")
            if isinstance(output_path, str):
                path = os.path.join(output_path, name)
                result.append(path)
                self.download_dataset(dataset_name=name, local_path=path)

        return result

    def get_memory_usage(self):
        header = self.__check_authentication()
        url = self.address + "/getMemoryUsage"
        response = requests.get(url, headers=header)
        if response.status_code != 200:
            raise ValueError("Code {}. {}".format(response.status_code, response.json().get("error")))
        response = response.json()
        info_list = response.get("infoList")
        index = []
        values = []
        for d in info_list:
            key = d['key']
            index.append(key)
            value = d['value']
            values.append(value)
        return pd.Series(data=values, index=index)


    """
        Execution
    """

    def trace_job(self, jobId):
        """ Get information about the specified remote job

        :param jobId: the job identifier
        :return: a dictionary with the information
        """
        header = self.__check_authentication()
        status_url = self.address + "/jobs/" + jobId + "/trace"
        status_resp = requests.get(status_url, headers=header)
        if status_resp.status_code != 200:
            raise ValueError("Code {}. {}".format(status_resp.status_code, status_resp.json().get("error")))
        return status_resp.json()


def create_callback(encoder, n_files=None):
    encoder_len = encoder.len
    if n_files is not None:
        tot_len = n_files
        byte_per_file = encoder_len / n_files
    else:
        tot_len = encoder_len

    if is_progress_enabled():
        bar = tqdm(total=tot_len)

        if n_files is not None:
            def callback(monitor):
                bar.update(max(int((monitor.bytes_read / byte_per_file) - bar.n), 0))

            return callback

        def callback(monitor):
            bar.update(monitor.bytes_read - bar.n)

        return callback
    else:
        def callback(monitor):
            pass
        return callback





import pandas as pd

LOCAL = "local"
REMOTE = "remote"
PARSER = "parser"


class SourcesTable:
    """ Handles the location of the datasets during the program execution.
    Manages both local and remote datasets. It is used when a mode change is
    detected to send or download data sources to/from the remote server.
    """
    def __init__(self):
        self.table = pd.DataFrame(columns=[LOCAL, REMOTE, PARSER])
        self.id_count = 0

    def add_source(self, local=None, remote=None, parser=None):
        self.table.loc[self.id_count] = {LOCAL: local, REMOTE: remote, PARSER: parser}
        result = self.id_count
        self.id_count += 1
        return result

    def get_source(self, id):
        res = self.table.loc[id].to_dict()
        return res

    def modify_source(self, id, local=None, remote=None):
        if local is not None:
            self.table.loc[id][LOCAL] = local
        if remote is not None:
            self.table.loc[id][REMOTE] = remote

    def search_source(self, local=None, remote=None):
        if local is not None:
            ids = self.table[self.table[LOCAL] == local].index.tolist()
        elif remote is not None:
            ids = self.table[self.table[REMOTE] == remote].index.tolist()
        else:
            raise ValueError("local or remote must be given")
        return ids[0] if len(ids) > 0 else None



from .MetaField import MetaField
from .RegField import RegField


def SQRT(argument):
    """ Computes the square matrix of the argument

    :param argument: a dataset region field (dataset.field) or metadata (dataset['field'])
    """
    if isinstance(argument, MetaField):
        return argument._unary_expression("SQRT")
    elif isinstance(argument, RegField):
        return argument._unary_expression("SQRT")
    else:
        raise TypeError("You have to give as input a RegField (dataset.field)"
                        "or a MetaField (dataset['field']")




from ..parsers.RegionParser import RegionParser
from . import *
from ..parsers import get_type_name


def parserToXML(parser, datasetName, path):
    if not isinstance(parser, RegionParser):
        raise TypeError("parser must be a Parser. {} was given".format(type(parser)))

    result = schema_header
    result += schema_collection_template.format(datasetName)
    result += schema_dataset_type_template.format(parser.get_parser_type(), parser.get_coordinates_system())

    for f, t in zip(parser.get_ordered_attributes(), map(get_type_name, parser.get_ordered_types())):
        result += schema_field_template.format(t, f)

    result += schema_dataset_end
    result += schema_collection_end

    with open(path, "w") as f:
        f.write(result)



from glob import glob
import tqdm
import os
import xml.etree.ElementTree as ET
from ..parsers.RegionParser import RegionParser
from ..parsers import GTF, COORDS_DEFAULT
import logging
import pandas as pd
from . import generateNameKey
from ..DataStructures import reg_fixed_fileds, \
    chr_aliases, start_aliases, stop_aliases, strand_aliases
import numpy as np
from ... import is_progress_enabled


# global logger
logger = logging.getLogger("PyGML logger")


def load_reg_from_path(path, parser=None):
    if parser is None:
        # get the parser for the dataset
        parser = get_parser(path)
    # we need to take only the files of the regions, so only the files that does NOT end with '.meta'
    all_files = set(glob(pathname=path + '/*'))
    meta_files = set(glob(pathname=path + '/*.meta'))

    only_region_files = all_files - meta_files
    logger.info("Loading region data from path {}".format(path))

    n_files = len(only_region_files)
    dfs = []
    for file in tqdm.tqdm(only_region_files, total=n_files, disable=not is_progress_enabled()):
        if file.endswith("schema") or file.endswith("_SUCCESS") or \
           file.endswith(".xml"):
            continue
        abs_path = os.path.abspath(file)
        key = generateNameKey(abs_path)

        df = parser.parse_regions(abs_path)
        df.index = np.repeat(key, len(df))
        dfs.append(df)
    if len(dfs) > 0:
        result = pd.concat(objs=dfs, copy=False)
    else:
        result = pd.DataFrame(columns=parser.get_ordered_attributes())
        result.index.name = "id_sample"
    # ordering the columns
    new_reg_fixed_fields = reg_fixed_fileds.copy()
    if parser.strandPos is None:
        new_reg_fixed_fields.remove("strand")
    result = result[new_reg_fixed_fields + [c for c in result.columns if c not in new_reg_fixed_fields]]
    # sorting index
    result = result.sort_index()
    return result


def get_parser(path):
    schema_file = get_schema_path(path)
    tree = ET.parse(schema_file)
    gmqlSchema = tree.getroot().getchildren()[0]
    parser_type = gmqlSchema.get('type')
    coordinate_system = gmqlSchema.get("coordinate_system")
    if coordinate_system is None:
        coordinate_system = COORDS_DEFAULT
    field_nodes = gmqlSchema.getchildren()

    i = 0
    chrPos, startPos, stopPos, strandPos, otherPos = None, None, None, None, None
    otherPos = []

    if parser_type == GTF:
        chrPos = 0      # seqname
        startPos = 3    # start
        stopPos = 4     # end
        strandPos = 6   # strand
        otherPos = [(1, 'source', 'string'), (2, 'feature', 'string'),
                    (5, 'score', 'float'), (7, 'frame', 'string')]

        for field in field_nodes:
            name = list(field.itertext())[0].lower()
            type = field.get('type').lower()
            if name not in {'seqname', 'start', 'end', 'strand',
                            'source', 'feature', 'score', 'frame'}:
                otherPos.append((i, name, type))
            i += 1

    else:
        for field in field_nodes:
            name = list(field.itertext())[0].lower()
            type = field.get('type').lower()

            if name in chr_aliases and chrPos is None:
                chrPos = i
            elif name in start_aliases and startPos is None:
                startPos = i
            elif name in stop_aliases and stopPos is None:
                stopPos = i
            elif name in strand_aliases and strandPos is None:
                strandPos = i
            else:   # other positions
                otherPos.append((i, name, type))
            i += 1
    if len(otherPos) == 0:
        otherPos = None
    return RegionParser(chrPos, startPos, stopPos, strandPos,
                        otherPos, coordinate_system=coordinate_system,
                        schema_format=parser_type)


def get_schema_path(path):
    schema_paths = (glob(pathname=path + '/*.schema') + glob(pathname=path + "/schema.xml"))
    return schema_paths[0] if len(schema_paths) > 0 else None



from ... import get_python_manager, none, Some, _get_gateway
from . import coordinate_systems, get_parsing_function, null_values, GTF
import numpy as np
import pandas as pd
import re


class RegionParser:
    def __init__(self, chrPos, startPos, stopPos,
                 strandPos=None, otherPos=None, delimiter="\t",
                 coordinate_system='0-based', schema_format="del",
                 parser_name="parser"):
        """ Creates a custom region dataset

        :param chrPos: position of the chromosome column
        :param startPos: position of the start column
        :param stopPos: position of the stop column
        :param strandPos: (optional) position of the strand column. Default is None
        :param otherPos: (optional) list of tuples of the type [(pos, attr_name, typeFun), ...]. Default is None
        :param delimiter: (optional) delimiter of the columns of the file. Default "\t"
        :param coordinate_system: can be {'0-based', '1-based', 'default'}. Default is '0-based'
        :param schema_format: (optional) type of file. Can be {'tab', 'gtf', 'vcf', 'del'}. Default is 'del'
        :param parser_name: (optional) name of the parser. Default is 'parser'
        """
        if isinstance(delimiter, str):
            self.delimiter = delimiter
        if isinstance(chrPos, int) and chrPos >= 0:
            self.chrPos = chrPos
        else:
            raise ValueError("Chromosome position must be >=0")
        if isinstance(startPos, int) and startPos >= 0:
            self.startPos = startPos
        else:
            raise ValueError("Start position must be >=0")
        if isinstance(stopPos, int) and stopPos >= 0:
            self.stopPos = stopPos
        else:
            raise ValueError("Stop position must be >=0")
        if isinstance(strandPos, int) and strandPos >= 0:
            self.strandPos = strandPos
            strandGmql = Some(strandPos)
        elif strandPos is None:
            self.strandPos = None
            strandGmql = none()
        else:
            raise ValueError("Strand position must be >= 0")
        if isinstance(parser_name, str):
            self.parser_name = parser_name
        else:
            raise TypeError("Parser name must be a string")
        if isinstance(coordinate_system, str):
            if coordinate_system in coordinate_systems:
                self.coordinate_system = coordinate_system
            else:
                raise ValueError("{} is not a valid coordinate system".format(coordinate_system))
        else:
            raise TypeError("Coordinate system must be a string")
        if isinstance(schema_format, str):
            self.schema_format = schema_format
        else:
            raise TypeError("Schema Format must be a string")

        if isinstance(otherPos, list):
            self.otherPos = convert_otherPos(otherPos)
            otherPosGmql = Some(convert_to_gmql(otherPos))
        else:
            otherPosGmql = none()
            self.otherPos = None

        pmg = get_python_manager()

        self.gmql_parser = pmg.buildParser(delimiter, self.chrPos, self.startPos,
                                           self.stopPos, strandGmql, otherPosGmql,
                                           self.schema_format, self.coordinate_system)

    def get_coordinates_system(self):
        return self.coordinate_system

    def get_parser_type(self):
        return self.schema_format

    def get_gmql_parser(self):
        """ Gets the Scala implementation of the parser

        :return: a Java Object
        """
        return self.gmql_parser

    def parse_strand(self, strand):
        """ Defines how to parse the strand column

        :param strand: a string representing the strand
        :return: the parsed result
        """
        if strand in ['+', '-', '*']:
            return strand
        else:
            return '*'

    def parse_regions(self, path):
        """ Given a file path, it loads it into memory as a Pandas dataframe

        :param path: file path
        :return: a Pandas Dataframe
        """
        if self.schema_format.lower() == GTF.lower():
            res = self._parse_gtf_regions(path)
        else:
            res = self._parse_tab_regions(path)
        return res

    def _parse_tab_regions(self, path):
        types = self.get_name_type_dict()
        if "strand" in types.keys():
            types.pop("strand")
        df = pd.read_csv(filepath_or_buffer=open(path), na_values=null_values,
                         header=None,
                         names=self.get_ordered_attributes(),
                         dtype=types,
                         sep=self.delimiter,
                         converters={'strand': self.parse_strand})
        return df

    def _parse_gtf_regions(self, path):
        def split_attributes(attrs):
            res = {}
            attr_splits = re.split(";\s*", attrs)
            for a in attr_splits:
                if len(a) > 0:
                    splits = re.split("\"\s*", a)
                    #             print(splits)
                    attr_name = splits[0].strip()
                    attr_value = splits[1].strip()
                    res[attr_name] = attr_value
            return res
        actual_attributes = self.get_ordered_attributes()[:8] + ['attributes']
        types = self.get_name_type_dict()
        types.pop("strand")
        df = pd.read_csv(filepath_or_buffer=open(path), sep=self.delimiter,
                         na_values=null_values,
                         names=actual_attributes,
                         dtype=types,
                         converters={'strand': self.parse_strand})
        df = pd.concat([df, pd.DataFrame(df.attributes.map(split_attributes).tolist())], axis=1)\
            .drop("attributes", axis=1)
        return df

    def get_attributes(self):
        """ Returns the unordered list of attributes

        :return: list of strings
        """
        attr = ['chr', 'start', 'stop']
        if self.strandPos is not None:
            attr.append('strand')
        if self.otherPos:
            for i, o in enumerate(self.otherPos):
                attr.append(o[1])

        return attr

    def get_ordered_attributes(self):
        """ Returns the ordered list of attributes

        :return: list of strings
        """
        attrs = self.get_attributes()
        attr_arr = np.array(attrs)
        poss = [self.chrPos, self.startPos, self.stopPos]
        if self.strandPos is not None:
            poss.append(self.strandPos)
        if self.otherPos:
            for o in self.otherPos:
                poss.append(o[0])

        idx_sort = np.array(poss).argsort()
        return attr_arr[idx_sort].tolist()

    def get_types(self):
        """ Returns the unordered list of data types

        :return: list of data types
        """
        types = [str, int, int]
        if self.strandPos is not None:
            types.append(str)
        if self.otherPos:
            for o in self.otherPos:
                types.append(o[2])

        return types

    def get_name_type_dict(self):
        """ Returns a dictionary of the type
        {'column_name': data_type, ...}

        :return: dict
        """
        attrs = self.get_attributes()
        types = self.get_types()
        d = dict()
        for i,a in enumerate(attrs):
            d[a] = types[i]

        return d

    def get_ordered_types(self):
        """ Returns the ordered list of data types

        :return: list of data types
        """
        types = self.get_types()
        types_arr = np.array(types)
        poss = [self.chrPos, self.startPos, self.stopPos]
        if self.strandPos is not None:
            poss.append(self.strandPos)
        if self.otherPos:
            for o in self.otherPos:
                poss.append(o[0])
        idx_sort = np.array(poss).argsort()
        return types_arr[idx_sort].tolist()


def convert_otherPos(otherPos):
    # print(otherPos)
    return list(map(_to_parsing_function, otherPos))


def _to_parsing_function(tpos):
    if len(tpos) != 3:
        raise ValueError("Position tuple has wrong number of parameters")
    else:
        if isinstance(tpos[2], str):
            fun = get_parsing_function(tpos[2].lower())
        else:
            TypeError("Type of region field must be a string")
        if isinstance(tpos[0], int) and isinstance(tpos[1], str):
            return tpos[0], tpos[1], fun
        else:
            raise TypeError("Position must be integer and name of field must be a string")


def convert_to_gmql(otherPos):
    otherPosJavaList = _get_gateway().jvm.java.util.ArrayList()
    for tpos in otherPos:
        posJavaList = _get_gateway().jvm.java.util.ArrayList()
        if len(tpos) != 3:
            raise ValueError("Position tuple has wrong number of parameters")
        else:
            if isinstance(tpos[0], int) and tpos[0] >= 0:
                posJavaList.append(str(tpos[0]))
            else:
                raise ValueError("Position must be >= 0")
            if isinstance(tpos[1], str):
                posJavaList.append(tpos[1])
            else:
                raise TypeError("Name of the field must be a string")
            if isinstance(tpos[2], str):
                get_parsing_function(tpos[2])
                posJavaList.append(tpos[2])
        otherPosJavaList.append(posJavaList)
    return otherPosJavaList



#!/usr/bin/env python

import gmql as gl
import argparse
import getpass
import time


def login():
    address = input("http address of web service: ")
    username = input("username: ")
    password = getpass.getpass("password: ")

    gl.set_remote_address(address)
    try:
        rm = gl.RemoteManager(address=address)
        rm.login(username=username, password=password)
        auth_token = rm.auth_token
        s_type = "authenticated"
        s_time = int(time.time())
        return rm, address, auth_token, s_type, s_time
    except Exception:
        print("Something went wrong. Try again.")
        exit(-1)


def main():
    parser = argparse.ArgumentParser(description='PyGMQL login parser')
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--login", help="insert the credential for the GMQL web service", action="store_true")
    group.add_argument("--logout", help="remove from the configuration file all the information for login",
                       action="store_true")

    args = parser.parse_args()
    if args.login:
        print("----- LOGIN TO GMQL WEB SERVICE -----")
        rm, address, auth_token, s_type, s_time = login()
        sm = gl.get_session_manager()
        sm.add_session(address, auth_token, s_time, s_type)

    if args.logout:
        print("----- LOGOUT FROM GMQL WEB SERVICE -----")
        rm, address, auth_token, s_type, s_time = login()
        sm = gl.get_session_manager()
        sm.remove_session(address)


if __name__ == '__main__':
    main()



import os

"""
    Generation of the index of the pandas dataframe.
    This can be done in different ways:
    - hashing the complete file name
    - using directly the file as index (this is visually appealing :) )
"""


def generateHashKey(filename):
    return hash(filename)


def generateNameKey(filename):
    filename = os.path.basename(filename)
    if filename.endswith(".meta"):
        return filename[:-5]
    else:
        return filename



from glob import glob
from tqdm import tqdm
import logging

from ..parsers.MetadataParser import GenericMetaParser
from . import generateNameKey
import os
import pandas as pd

logger = logging.getLogger("PyGML logger")


def load_meta_from_path(path):
    meta_files = glob(pathname=path + '/*.meta')
    parsed = []
    parser = GenericMetaParser()
    logger.info("Loading meta data from path {}".format(path))
    from ... import __disable_progress
    for f in tqdm(meta_files, total=len(meta_files), disable=__disable_progress):
        abs_path = os.path.abspath(f)
        abs_path_no_meta = abs_path[:-5]
        key = generateNameKey(abs_path_no_meta)
        ps = parser.parse_metadata(abs_path)                # [(attr_name, value), ...]
        ps = list(map(lambda x: (key, (x[0], x[1])), ps))   # [(id, (attr_name, value)),...]
        # parsing
        parsed.extend(ps)
    return to_pandas(parsed)


def to_pandas(meta_list):
    # turn to dictionary
    if len(meta_list) > 0:
        meta_list = list(map(to_dictionary, meta_list))     # [{'id_sample': id, attr_name: value},...]
        df = pd.DataFrame.from_dict(meta_list)
        columns = df.columns

        # grouping by 'id_sample'
        g = df.groupby('id_sample')

        logger.info("dataframe construction")
        result_df = pd.DataFrame()
        from ... import __disable_progress
        for col in tqdm(columns, total=len(columns), disable=__disable_progress):
            if col != 'id_sample':
                result_df[col] = g[col].apply(to_list)
    else:
        result_df = pd.DataFrame()
        result_df.index.name = "id_sample"
    return result_df


def to_dictionary(tuple):
    return {"id_sample": tuple[0], tuple[1][0]: tuple[1][1]}


def to_list(x):
    l = list(x)
    l = [a for a in l if not pd.isnull(a)]
    return l


class Aggregate(object):
    def __init__(self, name, argument=None):
        self.argument = argument
        self.name = name

    def get_aggregate_name(self):
        return self.name

    def get_argument(self):
        return self.argument

    def is_nullary(self):
        return self.argument is None

    def is_unary(self):
        return self.argument is not None


class COUNT(Aggregate):
    """ Counts the number of regions in the group. It is automatically computed by the
    :func:`~gmql.dataset.GMQLDataset.GMQLDataset.map` operation
    """
    def __init__(self):
        super(COUNT, self).__init__("COUNT")


class COUNTSAMP(Aggregate):
    """ Counts the number of samples in the group.
    """
    def __init__(self):
        super(COUNTSAMP, self).__init__("COUNTSAMP")


class SUM(Aggregate):
    """ Computes the sum of the values of the specified attribute
    """
    def __init__(self, argument):
        super(SUM, self).__init__("SUM", argument)


class MIN(Aggregate):
    """ Gets the minimum value in the aggregation group for the specified attribute
    """
    def __init__(self, argument):
        super(MIN, self).__init__("MIN", argument)


class MAX(Aggregate):
    """ Gets the maximum value in the aggregation group for the specified attribute
    """
    def __init__(self, argument):
        super(MAX, self).__init__("MAX", argument)


class AVG(Aggregate):
    """ Gets the average value in the aggregation group for the specified attribute
    """
    def __init__(self, argument):
        super(AVG, self).__init__("AVG", argument)


class BAG(Aggregate):
    """ Creates space-separated string of attribute values for the specified attribute.
    It is applicable to attributes of any type.
    """
    def __init__(self, argument):
        super(BAG, self).__init__("BAG", argument)


class BAGD(Aggregate):
    """ Creates space-separated string of DISTINCT attribute values for the specified attribute.
    It is applicable to attributes of any type.
    """
    def __init__(self, argument):
        super(BAGD, self).__init__("BAGD", argument)


class STD(Aggregate):
    """ Gets the standard deviation of the aggregation group for the specified attribute
    """
    def __init__(self, argument):
        super(STD, self).__init__("STD", argument)


class MEDIAN(Aggregate):
    """ Gets the median value of the aggregation group for the specified attribute
    """
    def __init__(self, argument):
        super(MEDIAN, self).__init__("MEDIAN", argument)


class Q1(Aggregate):
    """ Gets the first quartile for the specified attribute
    """
    def __init__(self, argument):
        super(Q1, self).__init__("Q1", argument)


class Q2(Aggregate):
    """ Gets the second quartile for the specified attribute
    """
    def __init__(self, argument):
        super(Q2, self).__init__("Q2", argument)


class Q3(Aggregate):
    """ Gets the third quartile for the specified attribute
    """
    def __init__(self, argument):
        super(Q3, self).__init__("Q3", argument)


# strings to recognize as NaN
NULL = "null"
INF = "∞"
UNDEFINED = "�"
null_values = [NULL, INF, UNDEFINED]

GTF = "gtf"
TAB = "tab"

string_aliases = ['string', 'char']
int_aliases = ['long', 'int', 'integer']
float_aliases = ['double', 'float']
allowed_types = ['bed', 'tab']


COORDS_ZERO_BASED = '0-based'
COORDS_ONE_BASED = '1-based'
COORDS_DEFAULT = 'default'
coordinate_systems = {COORDS_ZERO_BASED, COORDS_ONE_BASED, COORDS_DEFAULT}


def get_parsing_function(type_string):
    if type_string in string_aliases:
        return str
    elif type_string in int_aliases:
        return int
    elif type_string in float_aliases:
        return float
    else:
        raise ValueError("This type is not supported")


def get_type_name(type):
    if type == str:
        return "string"
    elif type == int:
        return "integer"
    elif type == float:
        return "double"
    else:
        raise ValueError("This type is not supported")


from .RegionParser import *
from .Parsers import *
from .MetadataParser import *


from ... import get_python_manager
from .MetaField import MetaField


class RegField:

    def __init__(self, name, index=None, region_condition=None, reNode=None):
        if (region_condition is not None) and (reNode is not None):
            raise ValueError("you cannot mix conditions and expressions")
        self.name = name
        self.index = index
        self.region_condition = region_condition
        pymg = get_python_manager()
        self.exp_build = pymg.getNewExpressionBuilder(self.index)
        # check that the name is not already complex
        if not (name.startswith("(") and name.endswith(")")) and reNode is None:
            self.reNode = self.exp_build.getRENode(name)
        else:
            self.reNode = reNode

    def is_complex(self):
        return self.name.startswith("(") and self.name.endswith(")")

    def getRegionCondition(self):
        return self.region_condition

    def getRegionExpression(self):
        return self.reNode

    """
        PREDICATES
    """

    def __eq__(self, other):
        return self._predicate(other, "EQ")

    def __ne__(self, other):
        return self._predicate(other, "NOTEQ")

    def __gt__(self, other):
        return self._predicate(other, "GT")

    def __ge__(self, other):
        return self._predicate(other, "GTE")

    def __lt__(self, other):
        return self._predicate(other, "LT")

    def __le__(self, other):
        return self._predicate(other, "LTE")

    def _predicate(self, other, operator):
        new_name = '(' + self.name + operator + str(other) + ')'
        other = self._process_other(other)
        predicate = self.exp_build.createRegionPredicate(self.name, operator, other)
        return RegField(name=new_name, region_condition=predicate, index=self.index)

    """
        CONDITIONS
    """

    def __and__(self, other):
        return self._binary_condition(other, "AND")

    def __or__(self, other):
        return self._binary_condition(other, "OR")

    def __invert__(self):
        return self._unary_condition("NOT")

    def _binary_condition(self, other, operator):
        if isinstance(other, RegField) and other.region_condition is not None:
            new_name = '(' + self.name + operator + other.name + ')'
            condition = self.exp_build.createRegionBinaryCondition(self.region_condition, operator, other.region_condition)
            return RegField(name=new_name, region_condition=condition, index=self.index)
        else:
            raise TypeError("Binary conditions can be done only between RegFields conditions")

    def _unary_condition(self, operator):
        new_name = operator + '(' + self.name + ')'
        condition = self.exp_build.createRegionUnaryCondition(self.region_condition, operator)
        return RegField(name=new_name, region_condition=condition, index=self.index)

    def isin(self, values):
        """ Selects the samples having the metadata attribute between the values provided
        as input

        :param values: a list of elements
        :return a new complex condition
        """
        if not isinstance(values, list):
            raise TypeError("Input should be a string. {} was provided".format(type(values)))
        if not (self.name.startswith("(") and self.name.endswith(")")):
            first = True
            new_condition = None
            for v in values:
                if first:
                    first = False
                    new_condition = self.__eq__(v)
                else:
                    new_condition = new_condition.__or__(self.__eq__(v))
            return new_condition
        else:
            raise SyntaxError("You cannot use 'isin' with a complex condition")

    """
        EXPRESSIONS
    """

    def __add__(self, other):
        return self._binary_expression(other, "ADD")

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self._binary_expression(other, "SUB", order="left")

    def __rsub__(self, other):
        return self._binary_expression(other, "SUB", order="right")

    def __mul__(self, other):
        return self._binary_expression(other, "MUL")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._binary_expression(other, "DIV", order="left")

    def __rtruediv__(self, other):
        return self._binary_expression(other, "DIV", order="right")

    def __neg__(self):
        return self._unary_expression("NEG")

    def __get_return_type(self, other):
        if isinstance(other, str):
            other_name = other
            other = self.exp_build.getREType("string", other)
        elif isinstance(other, int):
            other_name = str(other)
            other = self.exp_build.getREType("float", str(other))
        elif isinstance(other, float):
            other_name = str(other)
            other = self.exp_build.getREType("float", str(other))
        elif isinstance(other, RegField):
            other_name = other.name
            other = other.reNode
        elif isinstance(other, MetaField):
            if other.index != self.index:
                raise ValueError("You cannot mix fields or attributes of different"
                                 "dataset in a condition or expression")
            if other.reMetaNode is not None:
                other_name = other.name
                other = other.reMetaNode
            else:
                raise ValueError("You cannot use a complex Metadata Expression in a "
                                 "region expression. Only MetaFields are allowed.")
        else:
            raise ValueError("Expected string, float, integer, RegField or MetaField"
                             ". {} was found".format(type(other)))
        return other, other_name

    def _binary_expression(self, other, operation, order="left"):
        other, other_name = self.__get_return_type(other)
        if order == 'left':
            new_name = '(' + self.name + operation + other_name + ')'
            node = self.exp_build.getBinaryRegionExpression(self.reNode, operation, other)
        else:
            new_name = '(' + other_name + operation + self.name + ')'
            node = self.exp_build.getBinaryRegionExpression(other, operation, self.reNode)
        return RegField(name=new_name, index=self.index, reNode=node)

    def _unary_expression(self, operation):
        new_name = operation + "(" + self.name + ")"
        node = self.exp_build.getUnaryRegionExpression(self.reNode, operation)
        return RegField(name=new_name, index=self.index, reNode=node)

    def _process_other(self, other):
        if isinstance(other, MetaField):
            if other.index != self.index:
                raise ValueError("You cannot mix fields or attributes of different"
                                 "dataset in a condition or expression")
            if other.metaAccessor is not None:
                return other.metaAccessor
            else:
                raise ValueError("When mixing RegFields with metadata you can only use "
                                 "MetaFields of a metadata attribute and "
                                 "not complex conditions or expressions on metadata")
        elif isinstance(other, (str, int, float)):
            return str(other)
        else:
            raise TypeError("Other is of unknown type. "
                            "{} provided".format(type(other)))